"""
Assumptions:
Square voxels, no multi-frame support
"""

import os
import sys
import traceback
from copy import copy, deepcopy
from math import pi

import numpy as np
import scipy.optimize as opt
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement
from hazenlib.utils import Rod
from matplotlib import pyplot as plt
from scipy import ndimage
from scipy.interpolate import interp1d
from skimage.measure import regionprops


class SliceWidth(HazenTask):
    """Slice width measurement class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.single_dcm = self.dcm_list[0]
        self.pixel_size = self.single_dcm.PixelSpacing[0]

    def run(self):
        """Main function for performing slice width measurement

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        results = self.init_result_dict()
        results.files = self.img_desc(self.single_dcm)
        try:
            for measurement in self.get_slice_width(self.single_dcm):
                results.add_measurement(measurement)

        except Exception as e:
            logger.exception(
                "Could not calculate the slice_width for %s because of : %s",
                self.img_desc(self.single_dcm),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def sort_rods(self, rods):
        """Separate matrix of rods into sorted per row

        Args:
            rods (_type_): _description_

        Returns:
            _type_: _description_
        """
        lower_row = sorted(rods, key=lambda rod: rod.y)[-3:]
        lower_row = sorted(lower_row, key=lambda rod: rod.x)
        middle_row = sorted(rods, key=lambda rod: rod.y)[3:6]
        middle_row = sorted(middle_row, key=lambda rod: rod.x)
        upper_row = sorted(rods, key=lambda rod: rod.y)[0:3]
        upper_row = sorted(upper_row, key=lambda rod: rod.x)
        return lower_row + middle_row + upper_row

    def get_rods(self, arr):
        """Locate rods in the pixel array

        Args:
            arr (np.array): DICOM pixel array

        Returns:
            rods : array_like   centroid coordinates of rods
            rods_initial : array_like    initial guess at rods (center-of mass)

        Notes:
            The rod indices are ordered as:
                789
                456
                123
        """

        # inverted image for fitting (maximisation)
        arr_inv = np.invert(arr)
        if np.min(arr_inv) < 0:
            arr_inv = arr_inv + abs(
                np.min(arr_inv)
            )  # ensure voxel values positive for maximisation

        """ Initial Center-of-mass Rod Locator """

        # threshold and binaries the image in order to locate the rods.
        img_max = np.max(arr)  # maximum number of img intensity
        no_region = [None] * img_max

        img_tmp = arr
        # step over a range of threshold levels from 0 to the max in the image
        # using the ndimage.label function to count the features for each threshold
        for x in range(0, img_max):
            tmp = img_tmp <= x
            labeled_array, num_features = ndimage.label(tmp.astype(int))
            no_region[x] = num_features

        # find the indices that correspond to 10 regions and pick the median
        index = [i for i, val in enumerate(no_region) if val == 10]

        thres_ind = np.median(index).astype(int)

        # Generate the labelled array with the threshold chosen
        img_threshold = img_tmp <= thres_ind

        labeled_array, num_features = ndimage.label(img_threshold.astype(int))

        # check that we have got the 10 rods!
        if num_features != 10:
            sys.exit("Did not find the 9 rods")

        # list of tuples of x,y coordinates for the centres
        rod_centres = ndimage.center_of_mass(arr, labeled_array, range(2, 11))

        rods = [Rod(x=x[1], y=x[0]) for x in rod_centres]
        rods = self.sort_rods(rods)
        rods_initial = deepcopy(rods)  # save for later

        """ Gaussian 2D Rod Locator """

        # setup bounding box dict
        # TODO: make these into Rod class properties and functions
        # rather than loop over 9 each time
        bbox = {
            "x_start": [],
            "x_end": [],
            "y_start": [],
            "y_end": [],
        }

        # Get average radius and inverse intensity of rods
        rprops = regionprops(labeled_array, arr_inv)[1:]  # ignore first label

        # get relevant label properties: radius and intensity
        bbox["rod_dia"] = [prop.feret_diameter_max for prop in rprops]
        bbox["intensity_max"] = [prop.intensity_max for prop in rprops]

        # Calculate mean
        radius_of_rods_mean = int(np.mean(bbox["rod_dia"]))
        # inv_intensity_of_rods_mean = int(np.mean(inv_intensity_of_rods))
        bbox["radius"] = int(np.ceil((radius_of_rods_mean * 2) / 2))

        # array extend bounding box regions around rods by radius no. pixels
        for rprop in rprops:
            bbox["x_start"].append(rprop.bbox[0] - bbox["radius"])
            bbox["x_end"].append(rprop.bbox[2] + bbox["radius"])
            bbox["y_start"].append(rprop.bbox[1] - bbox["radius"])
            bbox["y_end"].append(rprop.bbox[3] + bbox["radius"])

            # print(f'Rod {idx}   Bounding Box, x: ({bbox["x_start"][-1]}, {bbox["x_end"][-1]}), y: ({bbox["y_start"][-1]}, {bbox["y_end"][-1]})')

        x0, y0, x0_im, y0_im = ([None] * 9 for i in range(4))

        for idx in range(9):
            cropped_data = arr_inv[
                bbox["x_start"][idx] : bbox["x_end"][idx],
                bbox["y_start"][idx] : bbox["y_end"][idx],
            ]
            x0_im[idx], y0_im[idx], x0[idx], y0[idx] = (
                self.fit_gauss_2d_to_rods(
                    cropped_data,
                    bbox["intensity_max"][idx],
                    bbox["rod_dia"][idx],
                    bbox["radius"],
                    bbox["x_start"][idx],
                    bbox["y_start"][idx],
                )
            )
            # note: flipped x/y
            rods[idx].x = y0_im[idx]
            rods[idx].y = x0_im[idx]

        rods = self.sort_rods(rods)

        # save figure
        if self.report:
            fig, axes = plt.subplots(1, 3, figsize=(45, 15))
            fig.tight_layout(pad=1)
            # center-of-mass (original method)
            axes[0].set_title("Initial Estimate")
            axes[0].imshow(arr, cmap="gray")
            for idx in range(9):
                axes[0].plot(rods_initial[idx].x, rods_initial[idx].y, "y.")
            # gauss 2D
            axes[1].set_title("2D Gaussian Fit")
            axes[1].imshow(arr, cmap="gray")
            for idx in range(9):
                axes[1].plot(rods[idx].x, rods[idx].y, "r.")
            # combined
            axes[2].set_title("Initial Estimate vs. 2D Gaussian Fit")
            axes[2].imshow(arr, cmap="gray")
            for idx in range(9):
                axes[2].plot(rods_initial[idx].x, rods_initial[idx].y, "y.")
                axes[2].plot(rods[idx].x, rods[idx].y, "r.")
            img_path = os.path.realpath(
                os.path.join(
                    self.report_path,
                    f"{self.img_desc(self.single_dcm)}_rod_centroids.png",
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return rods, rods_initial

    def plot_rods(self, ax, arr, rods, rods_initial):  # pragma: no cover
        """Plot rods and curve fit graphs

        Args:
            ax (matplotlib.pyplot.axis): image axis
            arr (dcm.pixelarray): pixel array (image of phantom)
            rods (_type_): _description_
            rods_initial (_type_): _description_

        Returns:
            matplotlib.pyplot.axis: _description_
        """
        ax.imshow(arr, cmap="gray")
        for idx, rod in enumerate(rods):
            # ax.plot(rods_initial[idx].x, rods_initial[idx].y, 'y.', markersize=2)  # center-of-mass method
            ax.plot(rod.x, rod.y, "r.", markersize=2)  # gauss 2D
            ax.scatter(
                x=rod.x + 5,
                y=rod.y - 5,
                marker=f"${idx + 1}$",
                s=30,
                linewidths=0.4,
                c="w",
            )

        ax.set_title("Rod Centroids")
        return ax

    def get_rod_distances(self, rods):
        """
        Calculates horizontal and vertical distances between adjacent rods in pixels

        Parameters
        ----------
        rods : array_like
            rod positions in pixels

        Returns
        -------
        horz_dist, vert_dist : array_like
            horizontal and vertical distances between rods in pixels

        """
        # TODO: move to be a function of the Rod class
        horz_dist = [
            np.sqrt(
                np.square((rods[2].y - rods[0].y))
                + np.square(rods[2].x - rods[0].x)
            ),
            np.sqrt(
                np.square((rods[5].y - rods[3].y))
                + np.square(rods[5].x - rods[3].x)
            ),
            np.sqrt(
                np.square((rods[8].y - rods[6].y))
                + np.square(rods[8].x - rods[6].x)
            ),
        ]

        vert_dist = [
            np.sqrt(
                np.square((rods[0].y - rods[6].y))
                + np.square(rods[0].x - rods[6].x)
            ),
            np.sqrt(
                np.square((rods[1].y - rods[7].y))
                + np.square(rods[1].x - rods[7].x)
            ),
            np.sqrt(
                np.square((rods[2].y - rods[8].y))
                + np.square(rods[2].x - rods[8].x)
            ),
        ]

        return horz_dist, vert_dist

    def get_rod_distortion_correction_coefficients(
        self, horizontal_distances
    ) -> dict:
        """
        Removes the effect of geometric distortion from the slice width measurement. Assumes that rod separation is
        120 mm.

        Args:
            horizontal_distances (list): horizontal distances between rods, in pixels

        Returns:
            dict: dictionary containing top and bottom distortion coefficients, in mm
        """
        # TODO: move to be a function of the Rod class

        coefficients = {
            "top": np.mean(horizontal_distances[1:3]) * self.pixel_size / 120,
            "bottom": np.mean(horizontal_distances[0:2])
            * self.pixel_size
            / 120,
        }
        coefficients = {
            "top": np.mean(horizontal_distances[1:3]) * self.pixel_size / 120,
            "bottom": np.mean(horizontal_distances[0:2])
            * self.pixel_size
            / 120,
        }

        return coefficients

    def get_rod_distortions(self, horz_dist, vert_dist):
        """

        Args:
            horz_dist (list): horizontal distances
            vert_dist (list): vertical distances

        Returns:
            tuple of float: horizontal and vertical distortion values, in mm
        """
        # TODO: move to be a function of the Rod class

        # calculate the horizontal and vertical distances
        horz_dist_mm = np.multiply(self.pixel_size, horz_dist)
        vert_dist_mm = np.multiply(self.pixel_size, vert_dist)

        horz_distortion = (
            100 * np.std(horz_dist_mm, ddof=1) / np.mean(horz_dist_mm)
        )  # ddof to match MATLAB std
        horz_distortion = (
            100 * np.std(horz_dist_mm, ddof=1) / np.mean(horz_dist_mm)
        )  # ddof to match MATLAB std
        vert_distortion = (
            100 * np.std(vert_dist_mm, ddof=1) / np.mean(vert_dist_mm)
        )
        return horz_distortion, vert_distortion

    def baseline_correction(self, profile, sample_spacing):
        """Calculates quadratic fit of the baseline and subtracts from profile

        Args:
            profile (list)
            sample_spacing (int)

        Returns:
            dict: of polynomial_coefficients, x_interpolated,
                baseline/polynomial_fit, baseline, baseline_interpolated,
                profile_interpolated, profile_corrected_interpolated

        """
        profile_width = len(profile)
        padding = 30
        outer_profile = np.concatenate(
            [profile[0:padding], profile[-padding:]]
        )
        # create the x axis for the outer profile
        x_left = np.arange(padding)
        x_right = np.arange(profile_width - padding, profile_width)
        x_outer = np.concatenate([x_left, x_right])

        # seconds order poly fit of the outer profile
        polynomial_coefficients = np.polyfit(x_outer, outer_profile, 2)
        polynomial_fit = np.poly1d(polynomial_coefficients)

        # use the poly fit to generate a quadratic curve with 0.25 space (high res)
        x_interp = np.arange(0, profile_width, sample_spacing)
        x = np.arange(0, profile_width)

        baseline_interp = polynomial_fit(x_interp)
        baseline = polynomial_fit(x)

        # Remove the baseline effects from the profiles
        profile_corrected = profile - baseline
        f = interp1d(x, profile_corrected, fill_value="extrapolate")
        profile_corrected_interp = f(x_interp)
        profile_interp = profile_corrected_interp + baseline_interp

        return {
            "f": polynomial_coefficients,
            "x_interpolated": x_interp,
            "baseline_fit": polynomial_fit,
            "baseline": baseline,
            "baseline_interpolated": baseline_interp,
            "profile_interpolated": profile_interp,
            "profile_corrected_interpolated": profile_corrected_interp,
        }
        return {
            "f": polynomial_coefficients,
            "x_interpolated": x_interp,
            "baseline_fit": polynomial_fit,
            "baseline": baseline,
            "baseline_interpolated": baseline_interp,
            "profile_interpolated": profile_interp,
            "profile_corrected_interpolated": profile_corrected_interp,
        }

    def gauss_2d(self, xy_tuple, A, x_0, y_0, sigma_x, sigma_y, theta, C):
        """
        Create 2D Gaussian
        Based on code by Siân Culley, UCL/KCL
        See also: https://en.wikipedia.org/wiki/Gaussian_function#Two-dimensional_Gaussian_function

        Parameters
        ----------
        xy_tuple : grid of x-y coordinates
        A : amplitude of 2D Gaussian
        x_0 / y_0 : centre of 2D Gaussian
        sigma_x / sigma_y : widths of 2D Gaussian
        theta : rotation of Gaussian
        C : background/intercept of 2D Gaussian

        Returns
        -------
        gauss : 1-D list of Gaussian intensities

        """
        (x, y) = xy_tuple
        x_0 = float(x_0)
        y_0 = float(y_0)

        cos_theta_2 = np.cos(theta) ** 2
        sin_theta_2 = np.sin(theta) ** 2
        sin_2_theta = np.sin(2 * theta)

        sigma_x_2 = sigma_x**2
        sigma_y_2 = sigma_y**2
        sigma_x_2 = sigma_x**2
        sigma_y_2 = sigma_y**2

        a = cos_theta_2 / (2 * sigma_x_2) + sin_theta_2 / (2 * sigma_y_2)
        b = -sin_2_theta / (4 * sigma_x_2) + sin_2_theta / (4 * sigma_y_2)
        c = sin_theta_2 / (2 * sigma_x_2) + cos_theta_2 / (2 * sigma_y_2)

        gauss = (
            A
            * np.exp(
                -(
                    a * (x - x_0) ** 2
                    + 2 * b * (x - x_0) * (y - y_0)
                    + c * (y - y_0) ** 2
                )
            )
            + C
        )
        gauss = (
            A
            * np.exp(
                -(
                    a * (x - x_0) ** 2
                    + 2 * b * (x - x_0) * (y - y_0)
                    + c * (y - y_0) ** 2
                )
            )
            + C
        )

        return gauss.ravel()

    def fit_gauss_2d_to_rods(
        self,
        cropped_data,
        gauss_amp,
        gauss_radius,
        box_radius,
        x_start,
        y_start,
    ):
        """
        Fit 2D Gaussian to Rods
        - Important:
        --- This uses a cropped region around a rod. If the cropped region is too large,
        such that it includes signal with intensity similar to the rods, the fitting may fail.
        --- This is a maximisation function, hence the rods should have higher signal than the surrounding region
        Based on code by Siân Culley, UCL/KCL

        Args:
            cropped_data (np.array): 2D array of magnitude voxels (nb: should be inverted if rods hypointense)
            gauss_amp (float/int): initial estimate of amplitude of 2D Gaussian
            gauss_radius (int): initial estimate of centre of 2D Gaussian
            box_radius (int): 'radius' of box around rod
            x_start / y_start (int, int): coordinates of bounding box in original non-cropped data

        Returns:
            tuple of 4 values corresponding to:
                x0_im / y0_im : rod centroid coordinates in dimensions of original image
                x0 / y0 : rod centroid coordinates in dimensions of cropped image
        """

        # get (x,y) coordinates for fitting
        indices = np.indices(cropped_data.shape)

        # estimate initial conditions for 2d gaussian fit
        dims_crop = cropped_data.shape
        h_crop = dims_crop[0]
        w_crop = dims_crop[1]

        A = gauss_amp  # np.max() # amp of Gaussian
        sigma = gauss_radius / 2  # radius of 2D Gaussian
        C = np.mean(
            [
                cropped_data[0, 0],
                cropped_data[h_crop - 1, 0],
                cropped_data[0, w_crop - 1],
                cropped_data[h_crop - 1, w_crop - 1],
            ]
        )  # background   np.min(outside of rod within cropped_data)
        C = np.mean(
            [
                cropped_data[0, 0],
                cropped_data[h_crop - 1, 0],
                cropped_data[0, w_crop - 1],
                cropped_data[h_crop - 1, w_crop - 1],
            ]
        )  # background   np.min(outside of rod within cropped_data)

        # print("A:", A)
        # print("box_radius:", box_radius)
        # print("sigma:", sigma)
        # print("C:", C, "\n")

        p0 = [A, box_radius, box_radius, sigma, sigma, 0, C]
        # print(f'initial conditions for 2d gaussian fitting: {p0}\n')

        # do 2d gaussian fit to data
        popt_single, pcov_single = opt.curve_fit(
            self.gauss_2d, indices, cropped_data.ravel(), p0=p0
        )

        A = popt_single[0]
        x0 = popt_single[1]
        y0 = popt_single[2]
        C = popt_single[6]

        # to get image coordinates need to add back on x_start and y_start
        x0_im = x0 + x_start
        y0_im = y0 + y_start

        return x0_im, y0_im, x0, y0

    def trapezoid(
        self,
        n_ramp,
        n_plateau,
        n_left_baseline,
        n_right_baseline,
        plateau_amplitude,
    ):
        """

        Args:
            n_ramp
            n_plateau
            n_left_baseline
            n_right_baseline
            plateau_amplitude

        Returns:
            tuple: trapezoid and fwmh
        """

        if n_left_baseline < 1:
            left_baseline = []
        else:
            left_baseline = np.zeros(n_left_baseline)

        if n_ramp < 1:
            left_ramp = []
            right_ramp = []
        else:
            left_ramp = np.linspace(0, plateau_amplitude, n_ramp)
            right_ramp = np.linspace(plateau_amplitude, 0, n_ramp)

        if n_plateau < 1:
            plateau = []
        else:
            plateau = plateau_amplitude * np.ones(n_plateau)

        if n_right_baseline < 1:
            right_baseline = []
        else:
            right_baseline = np.zeros(n_right_baseline)

            trap = np.concatenate(
                [left_baseline, left_ramp, plateau, right_ramp, right_baseline]
            )
            trap = np.concatenate(
                [left_baseline, left_ramp, plateau, right_ramp, right_baseline]
            )
            fwhm = n_plateau + n_ramp

        return trap, fwhm

    def get_ramp_profiles(self, image_array, rods) -> dict:
        """Find the central y-axis point for the top and bottom profiles
        done by finding the distance between the mid-distances of the central rods

        Args:
            image_array (dcm.pixelarray): pixel array from a DICOM image
            rods (list of Rods): list of rods with x,y coordinates

        Returns:
            dict: top and bottom ramp profiles
        """

        top_profile_vertical_centre = np.round(
            ((rods[3].y - rods[6].y) / 2) + rods[6].y
        ).astype(int)
        bottom_profile_vertical_centre = np.round(
            ((rods[0].y - rods[3].y) / 2) + rods[3].y
        ).astype(int)
        top_profile_vertical_centre = np.round(
            ((rods[3].y - rods[6].y) / 2) + rods[6].y
        ).astype(int)
        bottom_profile_vertical_centre = np.round(
            ((rods[0].y - rods[3].y) / 2) + rods[3].y
        ).astype(int)

        # Selected 20mm around the mid-distances and take the average to find the line profiles
        top_profile = image_array[
            (top_profile_vertical_centre - round(10 / self.pixel_size)) : (
                top_profile_vertical_centre + round(10 / self.pixel_size)
            ),
            int(rods[3].x) : int(rods[5].x),
        ]

        bottom_profile = image_array[
            (bottom_profile_vertical_centre - round(10 / self.pixel_size)) : (
                bottom_profile_vertical_centre + round(10 / self.pixel_size)
            ),
            int(rods[3].x) : int(rods[5].x),
        ]

        return {
            "top": top_profile,
            "bottom": bottom_profile,
            "top-centre": top_profile_vertical_centre,
            "bottom-centre": bottom_profile_vertical_centre,
        }

    def get_initial_trapezoid_fit_and_coefficients(
        self, profile, slice_thickness
    ):
        """

        Args:
            profile
            slice_thickness (int)

        Returns:
            tuple: of trapezoid_fit_initial and  trapezoid_fit_coefficients
        """

        n_plateau, n_ramp = None, None

        if slice_thickness == 3:
            # not sure where these magic numbers are from, I subtracted 1 from MATLAB numbers
            n_ramp = 7
            n_plateau = 32

        elif slice_thickness == 5:
            # not sure where these magic numbers are from, I subtracted 1 from MATLAB numbers
            n_ramp = 47
            n_plateau = 55

        trapezoid_centre = int(
            round(np.median(np.argwhere(profile < np.mean(profile))))
        )

        n_total = len(profile)
        n_left_baseline = int(
            trapezoid_centre - round(n_plateau / 2) - n_ramp - 1
        )
        n_right_baseline = n_total - n_left_baseline - 2 * n_ramp - n_plateau
        plateau_amplitude = np.percentile(profile, 5) - np.percentile(
            profile, 95
        )
        trapezoid_fit_coefficients = [
            n_ramp,
            n_plateau,
            n_left_baseline,
            n_right_baseline,
            plateau_amplitude,
        ]

        trapezoid_fit_initial, _ = self.trapezoid(
            n_ramp,
            n_plateau,
            n_left_baseline,
            n_right_baseline,
            plateau_amplitude,
        )

        return trapezoid_fit_initial, trapezoid_fit_coefficients

    def fit_trapezoid(self, profiles, slice_thickness):
        """

        Args:
            profile
            slice_thickness (int)

        Returns:
            tuple: of trapezoid_fit_initial and  trapezoid_fit_coefficients

        """
        (
            trapezoid_fit,
            trapezoid_fit_coefficients,
        ) = self.get_initial_trapezoid_fit_and_coefficients(
            profiles["profile_corrected_interpolated"], slice_thickness
        )

        x_interp = profiles["x_interpolated"]
        profile_interp = profiles["profile_interpolated"]
        baseline_interpolated = profiles["baseline_fit"](x_interp)
        baseline_fit_coefficients = profiles["baseline_fit"]
        baseline_fit_coefficients = [
            baseline_fit_coefficients.c[0],
            baseline_fit_coefficients.c[1],
            baseline_fit_coefficients.c[2],
        ]
        # sum squared differences
        current_error = sum(
            (
                profiles["profile_corrected_interpolated"]
                - (baseline_interpolated + trapezoid_fit)
            )
            ** 2
        )

        def get_error(base, trap):
            """Check if fit is improving"""
            trapezoid_fit_temp, _ = self.trapezoid(*trap)

            baseline_fit_temp = np.poly1d(base)(x_interp)

            sum_squared_difference = sum(
                (profile_interp - (baseline_fit_temp + trapezoid_fit_temp))
                ** 2
            )

            return sum_squared_difference

        cont = 1
        j = 0
        # Go through a series of changes to reduce error,
        # if error doesn't reduced in one entire loop then exit
        while cont == 1:
            j += 1
            cont = 0

            for i in range(14):
                baseline_fit_coefficients_temp = copy(
                    baseline_fit_coefficients
                )
                trapezoid_fit_coefficients_temp = copy(
                    trapezoid_fit_coefficients
                )

                if i == 0:
                    baseline_fit_coefficients_temp[0] = (
                        baseline_fit_coefficients_temp[0] - 0.0001
                    )

                elif i == 1:
                    baseline_fit_coefficients_temp[0] = (
                        baseline_fit_coefficients_temp[0] + 0.0001
                    )
                elif i == 2:
                    baseline_fit_coefficients_temp[1] = (
                        baseline_fit_coefficients_temp[1] - 0.001
                    )
                elif i == 3:
                    baseline_fit_coefficients_temp[1] = (
                        baseline_fit_coefficients_temp[1] + 0.001
                    )
                elif i == 4:
                    baseline_fit_coefficients_temp[2] = (
                        baseline_fit_coefficients_temp[2] - 0.1
                    )
                elif i == 5:
                    baseline_fit_coefficients_temp[2] = (
                        baseline_fit_coefficients_temp[2] + 0.1
                    )
                elif i == 6:  # Decrease the ramp width
                    trapezoid_fit_coefficients_temp[0] = (
                        trapezoid_fit_coefficients_temp[0] - 1
                    )
                    trapezoid_fit_coefficients_temp[2] = (
                        trapezoid_fit_coefficients_temp[2] + 1
                    )
                    trapezoid_fit_coefficients_temp[3] = (
                        trapezoid_fit_coefficients_temp[3] + 1
                    )
                elif i == 7:  # Increase the ramp width
                    trapezoid_fit_coefficients_temp[0] = (
                        trapezoid_fit_coefficients_temp[0] + 1
                    )
                    trapezoid_fit_coefficients_temp[2] = (
                        trapezoid_fit_coefficients_temp[2] - 1
                    )
                    trapezoid_fit_coefficients_temp[3] = (
                        trapezoid_fit_coefficients_temp[3] - 1
                    )
                elif i == 8:  # Decrease plateau width
                    trapezoid_fit_coefficients_temp[1] = (
                        trapezoid_fit_coefficients_temp[1] - 2
                    )
                    trapezoid_fit_coefficients_temp[2] = (
                        trapezoid_fit_coefficients_temp[2] + 1
                    )
                    trapezoid_fit_coefficients_temp[3] = (
                        trapezoid_fit_coefficients_temp[3] + 1
                    )

                elif i == 9:  # Increase plateau width
                    trapezoid_fit_coefficients_temp[1] = (
                        trapezoid_fit_coefficients_temp[1] + 2
                    )
                    trapezoid_fit_coefficients_temp[2] = (
                        trapezoid_fit_coefficients_temp[2] - 1
                    )
                    trapezoid_fit_coefficients_temp[3] = (
                        trapezoid_fit_coefficients_temp[3] - 1
                    )

                elif i == 10:  # Shift centre to the left
                    trapezoid_fit_coefficients_temp[2] = (
                        trapezoid_fit_coefficients_temp[2] - 1
                    )
                    trapezoid_fit_coefficients_temp[3] = (
                        trapezoid_fit_coefficients_temp[3] + 1
                    )

                elif i == 11:  # Shift centre to the right
                    trapezoid_fit_coefficients_temp[2] = (
                        trapezoid_fit_coefficients_temp[2] + 1
                    )
                    trapezoid_fit_coefficients_temp[3] = (
                        trapezoid_fit_coefficients_temp[3] - 1
                    )

                elif i == 12:  # Reduce amplitude
                    trapezoid_fit_coefficients_temp[4] = (
                        trapezoid_fit_coefficients_temp[4] - 0.1
                    )

                elif i == 13:  # Increase amplitude
                    trapezoid_fit_coefficients_temp[4] = (
                        trapezoid_fit_coefficients_temp[4] + 0.1
                    )

                new_error = get_error(
                    base=baseline_fit_coefficients_temp,
                    trap=trapezoid_fit_coefficients_temp,
                )

                if new_error < current_error:
                    cont = 1
                    if i > 6:
                        trapezoid_fit_coefficients = (
                            trapezoid_fit_coefficients_temp
                        )
                    else:
                        baseline_fit_coefficients = (
                            baseline_fit_coefficients_temp
                        )
                    current_error = new_error

        return trapezoid_fit_coefficients, baseline_fit_coefficients

    def get_slice_width(self, dcm) -> Measurement:
        """Calculates slice width using double wedge image

        Args:
            dcm (pydicom.FileDataset): DICOM image object

        Returns:
            dict: including
                - slice_width_mm (float):
                    calculated slice width (top, bottom, combined; various methods) in mm
                - horizontal_linearity_mm, vertical_linearity_mm (float):
                    calculated average rod distance in mm
                - horz_distortion_mm, vert_distortion_mm (float)
                    calculated rod distance distortion in mm
        """
        slice_width_mm = {"top": {}, "bottom": {}, "combined": {}}
        arr = dcm.pixel_array
        sample_spacing = 0.25

        rods, rods_initial = self.get_rods(arr)
        horz_distances, vert_distances = self.get_rod_distances(rods)
        horz_distortion_mm, vert_distortion_mm = self.get_rod_distortions(
            horz_distances, vert_distances
        )
        correction_coefficients_mm = (
            self.get_rod_distortion_correction_coefficients(
                horizontal_distances=horz_distances
            )
        )

        ramp_profiles = self.get_ramp_profiles(arr, rods)
        ramp_profiles_baseline_corrected = {
            "top": self.baseline_correction(
                np.mean(ramp_profiles["top"], axis=0), sample_spacing
            ),
            "bottom": self.baseline_correction(
                np.mean(ramp_profiles["bottom"], axis=0), sample_spacing
            ),
        }

        trapezoid_coefficients, baseline_coefficients = self.fit_trapezoid(
            ramp_profiles_baseline_corrected["top"], dcm.SliceThickness
        )
        top_trap, fwhm = self.trapezoid(*trapezoid_coefficients)

        slice_width_mm["top"]["default"] = (
            fwhm * sample_spacing * self.pixel_size * np.tan((11.3 * pi) / 180)
        )
        # Factor of 4 because interpolated by factor of four

        slice_width_mm["top"]["geometry_corrected"] = (
            slice_width_mm["top"]["default"]
            / correction_coefficients_mm["top"]
        )

        # AAPM method directly incorporating phantom tilt
        slice_width_mm["top"]["aapm"] = fwhm * sample_spacing * self.pixel_size

        # AAPM method directly incorporating phantom tilt and independent of geometric linearity
        slice_width_mm["top"]["aapm_corrected"] = (
            fwhm * sample_spacing * self.pixel_size
        ) / correction_coefficients_mm["top"]

        trapezoid_coefficients, baseline_coefficients = self.fit_trapezoid(
            ramp_profiles_baseline_corrected["bottom"], dcm.SliceThickness
        )
        bottom_trap, fwhm = self.trapezoid(*trapezoid_coefficients)

        slice_width_mm["bottom"]["default"] = (
            fwhm * sample_spacing * self.pixel_size * np.tan((11.3 * pi) / 180)
        )
        # Factor of 4 because interpolated by factor of four

        slice_width_mm["bottom"]["geometry_corrected"] = (
            slice_width_mm["bottom"]["default"]
            / correction_coefficients_mm["bottom"]
        )

        # AAPM method directly incorporating phantom tilt
        slice_width_mm["bottom"]["aapm"] = (
            fwhm * sample_spacing * self.pixel_size
        )

        # AAPM method directly incorporating phantom tilt and independent of geometric linearity
        slice_width_mm["bottom"]["aapm_corrected"] = (
            fwhm * sample_spacing * self.pixel_size
        ) / correction_coefficients_mm["bottom"]

        # Geometric mean of slice widths (pg 34 of IPEM Report 80)
        slice_width_mm["combined"]["default"] = (
            slice_width_mm["top"]["default"]
            * slice_width_mm["bottom"]["default"]
        ) ** 0.5
        slice_width_mm["combined"]["geometry_corrected"] = (
            slice_width_mm["top"]["geometry_corrected"]
            * slice_width_mm["bottom"]["geometry_corrected"]
        ) ** 0.5

        # AAPM method directly incorporating phantom tilt
        theta = (180.0 - 2.0 * 11.3) * pi / 180.0
        term1 = (np.cos(theta)) ** 2.0 * (
            slice_width_mm["bottom"]["aapm"] - slice_width_mm["top"]["aapm"]
        ) ** 2.0 + (
            4.0
            * slice_width_mm["bottom"]["aapm"]
            * slice_width_mm["top"]["aapm"]
        )
        term2 = (
            slice_width_mm["bottom"]["aapm"] + slice_width_mm["top"]["aapm"]
        ) * np.cos(theta)
        term3 = 2.0 * np.sin(theta)

        slice_width_mm["combined"]["aapm_tilt"] = (term1**0.5 + term2) / term3

        # AAPM method directly incorporating phantom tilt and independent of geometric linearity
        theta = (180.0 - 2.0 * 11.3) * pi / 180.0
        term1 = (np.cos(theta)) ** 2.0 * (
            slice_width_mm["bottom"]["aapm_corrected"]
            - slice_width_mm["top"]["aapm_corrected"]
        ) ** 2.0 + (
            4.0
            * slice_width_mm["bottom"]["aapm_corrected"]
            * slice_width_mm["top"]["aapm_corrected"]
        )
        term2 = (
            slice_width_mm["bottom"]["aapm_corrected"]
            + slice_width_mm["top"]["aapm_corrected"]
        ) * np.cos(theta)
        term3 = 2.0 * np.sin(theta)

        slice_width_mm["combined"]["aapm_tilt_corrected"] = (
            term1**0.5 + term2
        ) / term3

        # calculate linearity in mm from distances in pixels

        horizontal_linearity_mm = np.mean(horz_distances) * self.pixel_size
        vertical_linearity_mm = np.mean(vert_distances) * self.pixel_size

        # calculate horizontal and vertical distances in mm from distances in pixels, for output

        horz_distances_mm = [
            round(x * self.pixel_size, 3) for x in horz_distances
        ]

        vert_distances_mm = [
            round(x * self.pixel_size, 3) for x in vert_distances
        ]

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(
                6, 1, gridspec_kw={"height_ratios": [3, 1, 1, 1, 1, 1]}
            )
            fig.set_size_inches(6, 16)
            fig.tight_layout(pad=1)

            self.plot_rods(axes[0], arr, rods, rods_initial)

            axes[1].plot(
                np.mean(ramp_profiles["top"], axis=0), label="mean top profile"
            )
            axes[1].plot(
                ramp_profiles_baseline_corrected["top"]["baseline"],
                label="top profile baseline (interpolated)",
            )
            axes[1].legend()

            axes[2].plot(
                ramp_profiles_baseline_corrected["top"][
                    "profile_corrected_interpolated"
                ],
                label="corrected top profile",
            )
            axes[2].plot(top_trap, label="trapezoid fit")
            axes[2].legend()

            axes[3].plot(
                np.mean(ramp_profiles["bottom"], axis=0),
                label="mean bottom profile",
            )
            axes[3].plot(
                ramp_profiles_baseline_corrected["bottom"]["baseline"],
                label="bottom profile baseline (interpolated",
            )
            axes[3].legend()

            axes[4].plot(
                ramp_profiles_baseline_corrected["bottom"][
                    "profile_corrected_interpolated"
                ],
                label="corrected bottom profile",
            )
            axes[4].plot(bottom_trap, label="trapezoid fit")
            axes[4].legend()
            axes[5].axis("off")
            axes[5].table(
                cellText=[
                    [str(x) for x in horz_distances_mm]
                    + [str(np.around(horizontal_linearity_mm, 3))],
                    [str(x) for x in vert_distances_mm]
                    + [str(np.around(vertical_linearity_mm, 3))],
                ],
                rowLabels=["H-distances (S->I)", "V-distances (R->L)"],
                colLabels=["1", "2", "3", "mean/linearity"],
                colWidths=[0.15]
                * (len(horz_distances) + 1),  # plus one for linearity,
                rowLoc="center",
                loc="center",
            )

            img_path = os.path.realpath(
                os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        # print(f"Series Description: {dcm.SeriesDescription}\nWidth: {dcm.Rows}\nHeight: {dcm.Columns}\nSlice Thickness(
        # mm):" f"{dcm.SliceThickness}\nField of View (mm): {hazenlib.get_field_of_view(dcm)}\nbandwidth (Hz/Px) : {
        # dcm.PixelBandwidth}\n" f"TR  (ms) : {dcm.RepetitionTime}\nTE  (ms) : {dcm.EchoTime}\nFlip Angle  (deg) : {
        # dcm.FlipAngle}\n" f"Horizontal line bottom (mm): {horz_distances[0]}\nHorizontal line middle (mm): {
        # horz_distances[2]}\n" f"Horizontal line top (mm): {horz_distances[2]}\nHorizontal Linearity (mm): {np.mean(
        # horz_distances)}\n" f"Horizontal Distortion: {horz_distortion}\nVertical line left (mm): {vert_distances[0]}\n"
        # f"Vertical line middle (mm): {vert_distances[1]}\nVertical line right (mm): {vert_distances[2]}\n" f"Vertical
        # Linearity (mm): {np.mean(vert_distances)}\nVertical Distortion: {vert_distortion}\n" f"Slice width top (mm): {
        # slice_width['top']['default']}\n" f"Slice width bottom (mm): {slice_width['bottom']['default']}\nPhantom tilt (
        # deg): {phantom_tilt_deg}\n" f"Slice width AAPM geometry corrected (mm): {slice_width['combined'][
        # 'aapm_tilt_corrected']}")

        return [
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="slice width",
                value=round(
                    slice_width_mm["combined"]["aapm_tilt_corrected"],
                    2,
                ),
                unit="mm",
            ),
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="distortion",
                description="Vertical",
                value=round(vert_distortion_mm, 2),
                unit="mm",
            ),
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="distortion",
                description="Horizontal",
                value=round(horz_distortion_mm, 2),
                unit="mm",
            ),
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="linearity",
                value=round(vertical_linearity_mm, 2),
                unit="mm",
                description="Vertical",
            ),
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="linearity",
                value=round(horizontal_linearity_mm, 2),
                unit="mm",
                description="Horizontal",
            ),
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="distance",
                value=horz_distances_mm,
                unit="mm",
                description="Horizontal",
            ),
            Measurement(
                name="SliceWidth",
                type="measured",
                subtype="distance",
                value=vert_distances_mm,
                unit="mm",
                description="Vertical",
            ),
        ]



"""
Map of local SNR across a flood phantom slice.

Introduction
============

The SNR for each voxel in an image (of a flood phantom) is estimated as the SNR
of a ROI centred on that voxel following the single image SNR method of McCann
et al. [1]_. The SNR map can show variation in SNR caused by smoothing filters.
It also highlights small regions of low signal which could be caused by micro-
bubbles or foreign bodies in the phantom. These inhomogeneities can erroneously
reduce SNR measurements made by other methods.


Algorithm overview
==================
1. Apply boxcar smoothing to original image to create smooth image.
2. Create noise image by subtracting smooth image from original image.
3. Create image mask to remove background using e.g.
     skimage.filters.threshold_minimum
4. Calculate SNR using McCann's method and overlay ROIs on image.
5. Estimate local noise as standard deviation of pixel values in ROI centred on
    a pixel. Repeat for each pixel in the noise image.
6. Plot the local noise as a heat map.


References
===========
McCann, A. J., Workman, A., & McGrath, C. (2013). A quick and robust
method for measurement of signal-to-noise ratio in MRI. Physics in Medicine
& Biology, 58(11), 3775.
"""

import os

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import skimage.morphology
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement
from scipy import ndimage
from skimage import filters


class SNRMap(HazenTask):
    """Signal-to-noise ratio mapping class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.single_dcm = self.dcm_list[0]
        # Initialise variables
        self.kernel_size = 9
        self.roi_size = 20
        self.roi_distance = 40
        # ----
        # * Scale ROI distance to account for different image sizes.
        # * Pass kernel_size and roi_size parameters from command line.

    def run(self):
        """Main function for performing signal-to-noise ratio mapping
        Returns SNR parametric map on flood phantom DICOM file.

        Notes:
            Five square ROIs are created, one at the image centre, and four peripheral
            ROIs with their centres displaced at 45, 135, 225 and 315 degrees from the
            centre. Displays and saves a parametric map.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        results = self.init_result_dict()
        img_desc = self.img_desc(self.single_dcm)
        results.files = img_desc

        #  Create original, smoothed and noise images
        #  ==========================================
        original, smoothed, noise = self.smooth(
            dcm=self.single_dcm, kernel=self.kernel_size
        )

        """
            #  Note: access NumPy arrays by column then row. E.g.
            #
            #  t=np.array([[1,2,3],[4,5,6]])
            #  t
            #  Out[118]:
            #  array([[1, 2, 3],
            #         [4, 5, 6]])
            #
            #  t[1,0]
            #  Out[119]: 4 # not 2
            #
            #  Confusingly, patches (circles, rectangles) use traditional [x,y]
            #  positioning. To centre a circle on pixel [a,b], the circle must be
            #  centred on [b,a]. The function np.flip(coords) can help.
        """

        #  Calculate mask and identify ROIs
        #  =======================
        image_centre, roi_corners = self.get_rois(smoothed)

        #  Calculate SNR
        #  =============
        snr = self.calc_snr(original, noise, roi_corners)

        #  Generate local SNR parametric map
        #  =================================
        snr_map = self.calc_snr_map(original, noise)

        results.add_measurement(
            Measurement(
                name="SNR",
                type="measured",
                subtype="smoothing",
                value=round(snr, 2),
            ),
        )

        if self.report:
            #  Plot images
            #  ===========
            fig_detailed = self.plot_detailed(
                original,
                smoothed,
                noise,
                snr,
                snr_map,
                image_centre,
                roi_corners,
            )
            fig_summary = self.plot_summary(snr_map, original, roi_corners)

            #  Save images
            #  ===========
            detailed_image_path = os.path.join(
                self.report_path, f"{img_desc}_snr_map_detailed.png"
            )
            summary_image_path = os.path.join(
                self.report_path, f"{img_desc}_snr_map.png"
            )

            fig_detailed.savefig(detailed_image_path, dpi=300)
            fig_summary.savefig(summary_image_path, dpi=300)

            self.report_files.append(summary_image_path)
            self.report_files.append(detailed_image_path)

            results.add_report_image(self.report_files)

        return results

    def smooth(self, dcm, kernel: int = 9):
        """Create noise and smoothed images from original_image.

        Args:
            dcm (pydicom.Dataset): DICOM image object
            kernel (int): Kernel used for smoothing. Default is 9x9 boxcar.

        Returns:
            tuple of np.ndarray: original, smoothed and noise images (pixel array)
        """
        original_image = dcm.pixel_array.astype(float)

        #  Warn if not 256 x 256 image
        #  TODO scale distances for other image sizes
        if original_image.shape != (256, 256):
            logger.warning(
                "Expected image size (256, 256). Image size is %r."
                " Algorithm untested with these dimensions.",
                original_image.shape,
            )

        normalised_kernel = (
            skimage.morphology.square(kernel)
            / skimage.morphology.square(kernel).sum()
        )
        # kernel = kernel / kernel.sum()  # normalise kernel
        smooth_image = ndimage.filters.convolve(
            original_image, normalised_kernel
        )

        #  Alternative method 1: OpenCV.
        # smooth_image = cv2.blur(original_image, (kernel_size, kernel_size))

        #  Alternative method 2: scipy.ndimage.
        # kernel = np.ones([kernel_size, kernel_size], float)
        # kernel = kernel / kernel.sum() # normalise kernel
        # smooth_image = ndimage.filters.convolve(original_image, kernel)
        #  Note: filters.convolve and filters.correlate produce identical output
        #  for symetric kernels. Be careful with other kernels.

        noise_image = original_image - smooth_image

        return original_image, smooth_image, noise_image

    def get_rois(self, smooth_image):
        """Identify phantom and generate ROI locations.

        Args:
            smooth_image (np.ndarray): pixel array of the smoothed image

        Returns:
            tuple of image_centre (tuple), roi_corners (list of int)
        """

        # Threshold from smooth_image to reduce noise effects
        threshold = filters.threshold_minimum(smooth_image)
        self.mask = smooth_image > threshold

        #  Get centroid (=centre of mass for binary image) and convert to array
        image_centre = np.array(ndimage.measurements.center_of_mass(self.mask))
        logger.debug("image_centre = %r.", image_centre)

        #  Store corner of centre ROI, cast as int for indexing
        roi_corners = [np.rint(image_centre - self.roi_size / 2).astype(int)]

        #  Add corners of remaining ROIs
        roi_distance = self.roi_distance
        roi_corners.append(roi_corners[0] + [-roi_distance, -roi_distance])
        roi_corners.append(roi_corners[0] + [roi_distance, -roi_distance])
        roi_corners.append(roi_corners[0] + [-roi_distance, roi_distance])
        roi_corners.append(roi_corners[0] + [roi_distance, roi_distance])

        return image_centre, roi_corners

    def calc_snr(self, original_image, noise_image, roi_corners):
        """Calculate SNR from original_image and noise_image.

        Args:
            original_image (np.ndarray): original pixel array
            noise_image (np.ndarray): pixel array of the image noise
            roi_corners (list): list of tuples corresponding to coordinates of the ROI corners

        Returns:
            float: signal to noise ratio value
        """
        roi_signal = []
        roi_noise = []

        for [x, y] in roi_corners:
            roi_signal.append(
                original_image[
                    x : x + self.roi_size, y : y + self.roi_size
                ].mean()
            )
            roi_noise.append(
                noise_image[x : x + self.roi_size, y : y + self.roi_size].std(
                    ddof=1
                )
            )
            # Note: *.std(ddof=1) uses sample standard deviation, default ddof=0
            # uses population std dev. Not sure which is statistically correct,
            # but using ddof=1 for consistency with IDL code.

        roi_snr = np.array(roi_signal) / np.array(roi_noise)
        snr = roi_snr.mean()

        logger.debug(
            "ROIs signal=%r, noise=%r, snr=%r", roi_signal, roi_noise, roi_snr
        )

        return snr

    def calc_snr_map(self, original_image, noise_image):
        """Calculate SNR map from original_image and noise_image.

        Args:
            original_image (np.ndarray): original pixel array
            noise_image (np.ndarray): pixel array of the image noise

        Returns:
            snr_map
        """
        #  If you need a faster (less transparent) implementation, see:
        #  https://nickc1.github.io/python,/matlab/2016/05/17/Standard-Deviation-(Filters)-in-Matlab-and-Python.html

        noise_map = ndimage.filters.generic_filter(
            noise_image, lambda x: np.std(x, ddof=1), size=self.roi_size
        )
        signal_map = ndimage.filters.uniform_filter(
            original_image, size=self.roi_size
        )
        snr_map = signal_map / noise_map

        return snr_map

    def draw_roi_rectangles(self, roi_corners, ax):
        """Add ROI rectangle overlays to plot.

        Args:
            roi_corners (list): list of coordinates (col, row) of ROI corners
            ax (matplotlib.axes): diagram axes to visualise rectangles on

        Returns:
            None
                adds rectangle overlay to matplotlib axes

        """
        for corner in roi_corners:
            rect = patches.Rectangle(
                np.flip(corner),
                self.roi_size,
                self.roi_size,
                linewidth=1,
                edgecolor="r",
                facecolor="none",
            )
            ax.add_patch(rect)

    def plot_snr_map(self, snr_map, fig, ax):
        """Add SNR map to a figure axis.

        Args:
            snr_map (__type__): SNR map diagram to visualise
            fig (matplotlib.pyplot.fig): figure handle
            ax (matplotlib.axes): diagram axes to visualise rectangles on

        Returns
            None
                adds SNR map overlay to matplotlib axes on figure
        """
        para_im = ax.imshow(snr_map, cmap="viridis", vmin=0)
        cax = fig.add_axes(
            [
                ax.get_position().x1 + 0.01,
                ax.get_position().y0,
                0.02,
                ax.get_position().height,
            ]
        )
        plt.colorbar(para_im, cax=cax)
        ax.set_title("SNR map")

    def plot_detailed(
        self,
        original_image,
        smooth_image,
        noise_image,
        snr,
        snr_map,
        image_centre,
        roi_corners,
    ):
        """Create 4-image detailed SNR map plots

        Args:
            original_image (np.ndarray): original image pixel array
            smooth_image (np.ndarray): smoothed pixel array
            noise_image (np.ndarray): noise image pixel array
            snr (float): SNR value to add to the plot title
            snr_map (np.ndarray): _description_
            image_centre (tuple or list): coordinates of the image centre
            roi_corners (list of list): coordinates (col, row) of ROI corners

        Returns:
            matplotlib.figure.Figure: figure handle with plots
        """
        fig, axs = plt.subplots(
            1, 4, sharex=True, sharey=True, figsize=(8, 2.8)
        )
        fig.suptitle(
            "SNR = %.2f (file: %s)"
            % (snr, os.path.basename(self.single_dcm.filename))
        )
        axs[0].imshow(original_image, cmap="gray")
        axs[0].set_title("Magnitude Image")
        axs[1].imshow(smooth_image, cmap="gray")
        axs[1].contour(self.mask, colors="y")
        phantom_centre_marker = patches.Circle(
            np.flip(np.rint(image_centre).astype("int")), color="y"
        )
        axs[1].add_patch(phantom_centre_marker)
        axs[1].set_title("Smoothed")
        axs[2].imshow(noise_image, cmap="gray")
        axs[2].set_title("Noise")
        self.draw_roi_rectangles(roi_corners, axs[0])
        self.draw_roi_rectangles(roi_corners, axs[2])
        self.plot_snr_map(snr_map, fig, axs[3])
        for ax in axs:
            ax.axis("off")

        return fig

    def plot_summary(self, snr_map, original_image, roi_corners):
        """Create 2-image summary SNR map plot.

        Args:
            original_image (np.ndarray): original image pixel array
            snr_map (np.ndarray): _description_
            roi_corners (list of list): coordinates (col, row) of ROI corners

        Returns:
            matplotlib.figure.Figure: figure handle with plots
        """
        fig, axs = plt.subplots(
            1, 2, sharex=True, sharey=True, figsize=(6, 2.8)
        )
        axs[0].imshow(original_image, cmap="gray")
        axs[0].set_title("Magnitude Image")

        self.draw_roi_rectangles(roi_corners, axs[0])
        self.plot_snr_map(snr_map, fig, axs[1])
        for ax in axs:
            ax.axis("off")

        return fig



"""
ACR SNR

Calculates the SNR for slice 7 (the uniformity slice) of the ACR phantom.

This script utilises the smoothed subtraction method described in McCann 2013 [1], and a standard subtraction SNR.

Created by Neil Heraghty (Adapted by Yassine Azma, yassine.azma@rmh.nhs.uk)

09/01/2023

[1] McCann, A. J., Workman, A., & McGrath, C. (2013). A quick and robust
method for measurement of signal-to-noise ratio in MRI. Physics in Medicine
& Biology, 58(11), 3775.
"""

# Typing imports
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pydicom

# Python imports
import contextlib
import os
import sys
import traceback
from pathlib import Path
from typing import Any

# Module imports
import numpy as np
from hazenlib.utils import get_dicom_files
from scipy import ndimage

# Local imports
import hazenlib.utils
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result

# Local imports
from hazenlib.utils import dcmread


class ACRSNR(HazenTask):
    """Signal-to-noise ratio measurement class for DICOM images of the ACR phantom."""

    def __init__(
        self,
        subtract: Path | str | None = None,
        measured_slice_width: float | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise the Hazen ACR SNR Object."""
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)
        self.ACR_obj = ACRObject(self.dcm_list)

        # measured slice width is expected to be a floating point number
        self.measured_slice_width = measured_slice_width
        with contextlib.suppress(TypeError):
            self.measured_slice_width = float(measured_slice_width)

        # subtract is expected to be a path to a folder
        try:
            self.subtract = Path(subtract)
        except TypeError:
            self.subtract = None
        else:
            if not self.subtract.is_dir():
                self.subtract = None

    def run(self) -> Result:
        """Main function for performing SNR measurement using slice 7 from the ACR phantom image set. Performs either
        smoothing or subtraction method depending on user-provided input.

        Notes:
            Uses the smoothing method by default or the subtraction method if a second set of images are provided (using the --subtract option with dataset in a separate folder).

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation.
        """
        # Identify relevant slice
        snr_dcm = self.ACR_obj.slice_stack[6]
        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())

        # SINGLE METHOD (SMOOTHING)
        if self.subtract is None:
            logger.debug(
                "Calculating SNR on %s with the single (smoothing) method",
                self.img_desc(snr_dcm),
            )
            try:
                results.files = self.img_desc(snr_dcm)
                snr, normalised_snr = self.snr_by_smoothing(
                    snr_dcm, self.measured_slice_width
                )
                results.add_measurement(
                    Measurement(
                        name="SNR",
                        type="measured",
                        subtype="smoothing",
                        value=round(snr, 2),
                        visibility="intermediate",
                    ),
                )
                results.add_measurement(
                    Measurement(
                        name="SNR",
                        type="normalised",
                        subtype="smoothing",
                        value=round(normalised_snr, 2),
                    ),
                )

            except Exception as e:
                logger.exception(
                    "Could not calculate the SNR for %s because of : %s",
                    self.img_desc(snr_dcm),
                    e,
                )
                traceback.print_exc(file=sys.stdout)
        # SUBTRACTION METHOD
        else:
            # Use the shared helper to get all DICOM files from the directory provided
            filepaths = get_dicom_files(Path(self.subtract))
            data2 = [dcmread(dicom) for dicom in filepaths]
            snr_dcm2 = ACRObject(data2).slice_stack[6]

            logger.debug(
                "Calculating SNR on %s and %s with the subtraction method",
                self.img_desc(snr_dcm),
                self.img_desc(snr_dcm2),
            )

            results.files = [self.img_desc(snr_dcm), self.img_desc(snr_dcm2)]
            try:
                snr, normalised_snr = self.snr_by_subtraction(
                    snr_dcm, snr_dcm2, self.measured_slice_width
                )

                results.add_measurement(
                    Measurement(
                        name="SNR",
                        type="measured",
                        subtype="subtraction",
                        value=round(snr, 2),
                        visibility="intermediate",
                    ),
                )
                results.add_measurement(
                    Measurement(
                        name="SNR",
                        type="normalised",
                        subtype="subtraction",
                        value=round(normalised_snr, 2),
                    ),
                )

            except Exception as e:
                logger.exception(
                    "Could not calculate the SNR for %s and %s"
                    " because of : %s",
                    self.img_desc(snr_dcm),
                    self.img_desc(snr_dcm2),
                    e,
                )
                traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def get_normalised_snr_factor(
        self, dcm, measured_slice_width=None
    ) -> float:
        """Calculates the normalisation factor to be applied to the SNR in order to obtain the absolute SNR (ASNR). The
        normalisation factor depends on voxel size, bandwidth, number of averages and number of phase encoding steps.

        Args:
            dcm (pydicom.Dataset): DICOM image object
            measured_slice_width (float, optional): Provide the true slice width for the set of images. Defaults to None.

        Returns:
            float: normalisation factor.
        """
        dx, dy = hazenlib.utils.get_pixel_size(dcm)
        bandwidth = hazenlib.utils.get_bandwidth(dcm)
        TR = hazenlib.utils.get_TR(dcm)
        rows = hazenlib.utils.get_rows(dcm)
        columns = hazenlib.utils.get_columns(dcm)

        if measured_slice_width:
            slice_thickness = measured_slice_width
        else:
            slice_thickness = hazenlib.utils.get_slice_thickness(dcm)

        averages = hazenlib.utils.get_average(dcm)
        bandwidth_factor = np.sqrt((bandwidth * columns / 2) / 1000) / np.sqrt(
            30
        )
        voxel_factor = 1 / (0.001 * dx * dy * slice_thickness)

        normalised_snr_factor = (
            bandwidth_factor
            * voxel_factor
            * (1 / (np.sqrt(averages * rows * (TR / 1000))))
        )
        return normalised_snr_factor

    def filtered_image(self, dcm: pydicom.Dataset) -> np.array:
        """Apply filtering to a pixel array (image), as per the single image SNR method outlined in McCann et al, 2013.

        Notes:
            Performs a 2D convolution (for filtering images), using uniform_filter (SciPy function).

        Args:
            dcm (pydicom.Dataset): DICOM image object.

        Returns:
            np.array: pixel array of the filtered image.
        """
        a = dcm.pixel_array.astype("int")

        # filter size = 9, following MATLAB code and McCann 2013 paper for head coil, although note McCann 2013
        # recommends 25x25 for body coil.
        # TODO add coil options, same as with MagNet SNR
        filtered_array = ndimage.uniform_filter(a, 25, mode="constant")
        return filtered_array

    def get_noise_image(self, dcm: pydicom.Dataset) -> np.array:
        """Get a noise image when only one set of DICOM data is available.

        Notes:
            Separates the image noise by smoothing the pixel array and subtracting the smoothed pixel array from the
            original, leaving only the noise.

        Args:
            dcm (pydicom.Dataset): DICOM image object

        Returns:
            np.array: pixel array representing the image noise.
        """
        a = dcm.pixel_array.astype("int")

        # Convolve image with boxcar/uniform kernel
        imsmoothed = self.filtered_image(dcm)

        # Subtract smoothed array from original
        imnoise = a - imsmoothed

        return imnoise

    def get_roi_samples(
        self,
        ax,
        dcm: pydicom.Dataset or np.ndarray,
        centre_col: int,
        centre_row: int,
    ) -> list:
        """Takes the pixel array and divides it into several rectangular regions of interest (ROIs). If 'ax' is provided, then a plot of the ROIs is generated.

        Args:
            ax (matplotlib.pyplot.subplots): matplotlib axis for visualisation.
            dcm (pydicom.Dataset or np.ndarray): DICOM image object, or its pixel array.
            centre_col (int): x coordinate of the centre.
            centre_row (int): y coordinate of the centre.

        Returns:
            list of np.array: subsets of the original pixel array.
        """
        data = dcm if isinstance(dcm, np.ndarray) else dcm.pixel_array

        sample = [None] * 5
        # for array indexing: [row, column] format
        sample[0] = data[
            (centre_row - 10) : (centre_row + 10),
            (centre_col - 10) : (centre_col + 10),
        ]
        sample[1] = data[
            (centre_row - 50) : (centre_row - 30),
            (centre_col - 50) : (centre_col - 30),
        ]
        sample[2] = data[
            (centre_row + 30) : (centre_row + 50),
            (centre_col - 50) : (centre_col - 30),
        ]
        sample[3] = data[
            (centre_row - 50) : (centre_row - 30),
            (centre_col + 30) : (centre_col + 50),
        ]
        sample[4] = data[
            (centre_row + 30) : (centre_row + 50),
            (centre_col + 30) : (centre_col + 50),
        ]

        if ax:
            from matplotlib.collections import PatchCollection
            from matplotlib.patches import Rectangle

            # for patches: [column/x, row/y] format

            rects = [
                Rectangle((centre_col - 10, centre_row - 10), 20, 20),
                Rectangle((centre_col - 50, centre_row - 50), 20, 20),
                Rectangle((centre_col + 30, centre_row - 50), 20, 20),
                Rectangle((centre_col - 50, centre_row + 30), 20, 20),
                Rectangle((centre_col + 30, centre_row + 30), 20, 20),
            ]
            pc = PatchCollection(
                rects, edgecolors="red", facecolors="None", label="ROIs"
            )
            ax.add_collection(pc)

        return sample

    def snr_by_smoothing(
        self, dcm: pydicom.Dataset, measured_slice_width=None
    ) -> float:
        """Obtains a noise image using the single-image smoothing technique. Generates a ROI within the phantom region
        of the pixel array. Then measures the mean signal within the ROI on the original pixel array, and the standard
        deviation within the ROI on the noise image. Calculates SNR using these values and multiplies the SNR by the
        normalisation factor.

        Args:
            dcm (pydicom.Dataset): DICOM image object.
            measured_slice_width (float, optional): Provide the true slice width for the set of images. Defaults to None.

        Returns:
            float: normalised_snr.
        """
        (centre_x, centre_y), _ = self.ACR_obj.find_phantom_center(
            dcm.pixel_array, self.ACR_obj.dx, self.ACR_obj.dy
        )
        noise_img = self.get_noise_image(dcm)

        signal = [
            np.mean(roi)
            for roi in self.get_roi_samples(
                ax=None, dcm=dcm, centre_col=centre_x, centre_row=centre_y
            )
        ]

        noise = [
            np.std(roi, ddof=1)
            for roi in self.get_roi_samples(
                ax=None,
                dcm=noise_img,
                centre_col=centre_x,
                centre_row=centre_y,
            )
        ]
        # note no root_2 factor in noise for smoothed subtraction (one image) method, replicating Matlab approach and
        # McCann 2013

        snr = np.mean(np.divide(signal, noise))

        normalised_snr = snr * self.get_normalised_snr_factor(
            dcm, measured_slice_width
        )

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(2, 1)
            fig.set_size_inches(8, 16)
            fig.tight_layout(pad=4)

            axes[0].imshow(dcm.pixel_array)
            axes[0].scatter(centre_x, centre_y, c="red")
            axes[0].set_title("Centroid Location")

            axes[1].set_title("Smoothed Noise Image")
            axes[1].imshow(noise_img, cmap="gray")
            self.get_roi_samples(axes[1], dcm, centre_x, centre_y)

            img_path = os.path.realpath(
                os.path.join(
                    self.report_path, f"{self.img_desc(dcm)}_smoothing.png"
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return snr, normalised_snr

    def snr_by_subtraction(
        self,
        dcm1: pydicom.Dataset,
        dcm2: pydicom.Dataset,
        measured_slice_width=None,
    ) -> float:
        """Calculates signal to noise ratio using the two image subtraction method. Obtains a noise image by subtracting
        the two pixel arrays. Obtains ROIs within the phantom region of the pixel arrays. Calculates the mean within
        the ROI on one of the pixel arrays, and the standard deviation within the ROIs on the noise image.
        Calculates the SNR with these measurements and multiplies by the normalisation factor.

        Args:
            dcm1 (pydicom.Dataset): DICOM image object to calculate signal.
            dcm2 (pydicom.Dataset): DICOM image object to calculate noise.
            measured_slice_width (float, optional): Provide the true slice width for the set of images. Defaults to None.

        Returns:
            float: normalised_snr.
        """
        (centre_x, centre_y), _ = self.ACR_obj.find_phantom_center(
            dcm1.pixel_array, self.ACR_obj.dx, self.ACR_obj.dy
        )

        difference = np.subtract(
            dcm1.pixel_array.astype("int"), dcm2.pixel_array.astype("int")
        )

        signal = [
            np.mean(roi)
            for roi in self.get_roi_samples(
                ax=None, dcm=dcm1, centre_col=centre_x, centre_row=centre_y
            )
        ]
        noise = np.divide(
            [
                np.std(roi, ddof=1)
                for roi in self.get_roi_samples(
                    ax=None,
                    dcm=difference,
                    centre_col=centre_x,
                    centre_row=centre_y,
                )
            ],
            np.sqrt(2),
        )
        snr = np.mean(np.divide(signal, noise))

        normalised_snr = snr * self.get_normalised_snr_factor(
            dcm1, measured_slice_width
        )

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(2, 1)
            fig.set_size_inches(8, 16)
            fig.tight_layout(pad=4)

            axes[0].imshow(dcm1.pixel_array)
            axes[0].scatter(centre_x, centre_y, c="red")
            axes[0].axis("off")
            axes[0].set_title("Centroid Location")

            axes[1].set_title("Difference Image")
            axes[1].imshow(
                difference,
                cmap="gray",
            )
            self.get_roi_samples(axes[1], dcm1, centre_x, centre_y)
            axes[1].axis("off")

            img_path = os.path.realpath(
                os.path.join(
                    self.report_path,
                    f"{self.img_desc(dcm1)}_snr_subtraction.png",
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return snr, normalised_snr



import os
import sys
import traceback

import cv2 as cv
import numpy as np
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result
from hazenlib.utils import get_pe_direction, get_pixel_size, rescale_to_byte


class Ghosting(HazenTask):
    """Ghosting measurement class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.single_dcm = self.dcm_list[0]

    def run(self) -> Result:
        """Main function for performing ghosting measurement

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM SeriesDescription + EchoTime + NumberOfAverages, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        results = self.init_result_dict()
        img_desc = self.img_desc(
            self.single_dcm,
            properties=["SeriesDescription", "EchoTime", "NumberOfAverages"],
        )
        results.files = img_desc

        try:
            ghosting_value = self.get_ghosting(self.single_dcm)
            results.add_measurement(
                Measurement(
                    name="Ghosting",
                    type="measured",
                    unit="%",
                    value=round(ghosting_value, 3),
                ),
            )

        except Exception as e:
            logger.exception(
                "Could not calculate the ghosting for %s because of : %s",
                img_desc,
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def calculate_ghost_intensity(
        self, ghost: np.ndarray, phantom: np.ndarray, noise: np.ndarray
    ) -> float:
        """Calculates the ghost intensity

        References: IPEM Report 112 - Small Bottle Method
                    MagNET, Ghosting = (Sg-Sn)/(Sp-Sn) x 100%

        Args:
            ghost (np.ndarray):
            phantom (np.ndarray):
            noise (np.ndarray):

        Returns:
            float
        """

        if ghost is None or phantom is None or noise is None:
            raise Exception(
                "At least one of ghost, phantom and noise ROIs is empty or null"
            )

        ghost_mean = np.mean(ghost)
        phantom_mean = np.mean(phantom)
        noise_mean = np.mean(noise)

        if phantom_mean < ghost_mean or phantom_mean < noise_mean:
            raise Exception(
                "The mean phantom signal is lower than the ghost or the noise signal. This can't be the case "
            )

        return 100 * abs(ghost_mean - noise_mean) / phantom_mean

    def get_signal_bounding_box(self, array: np.ndarray):
        """_summary_

        Args:
            array (np.ndarray): _description_

        Returns:
            tuple: positions of left_column, right_column, upper_row, lower_row
        """
        max_signal = np.max(array)

        signal_limit = np.percentile(max_signal, 0.95) * 0.4
        signal = []
        for idx, voxel in np.ndenumerate(array):
            if voxel > signal_limit:
                signal.append(idx)

        signal_column = sorted([voxel[1] for voxel in signal])
        signal_row = sorted([voxel[0] for voxel in signal])

        upper_row = min(signal_row)
        lower_row = max(signal_row)
        left_column = min(signal_column)
        right_column = max(signal_column)
        return (
            left_column,
            right_column,
            upper_row,
            lower_row,
        )

    def get_signal_slice(self, bounding_box, slice_radius=5):
        """_summary_

        Args:
            bounding_box (tuple): left_column, right_column, upper_row, lower_row
            slice_radius (int, optional): _description_. Defaults to 5.

        Returns:
            tuple of np.array: _description_
        """
        left_column, right_column, upper_row, lower_row = bounding_box
        centre_row = (upper_row + lower_row) // 2
        centre_column = (left_column + right_column) // 2
        idxs = (
            np.array(
                range(
                    centre_column - slice_radius, centre_column + slice_radius
                ),
                dtype=np.intp,
            )[:, np.newaxis],
            np.array(
                range(centre_row - slice_radius, centre_row + slice_radius),
                dtype=np.intp,
            ),
        )
        return idxs

    def get_background_rois(self, dcm, signal_centre):
        """Create pixel arrays of the selected regions of interest from the background

        Args:
            dcm (pydicom.Dataset): DICOM image object
            signal_centre (list): x, y coordinates of the centre

        Returns:
            list: pixel arrays of the background regions of interest
        """
        background_rois = []

        if (
            get_pe_direction(dcm) == "ROW"
        ):  # phase encoding is left -right i.e. increases with columns
            if (
                signal_centre[1] < dcm.Rows * 0.5
            ):  # phantom is in top half of image
                background_rois_row = round(
                    dcm.Rows * 0.75
                )  # in the bottom quadrant
            else:  # phantom is bottom half of image
                background_rois_row = round(
                    dcm.Rows * 0.25
                )  # in the top quadrant
            background_rois.append((signal_centre[0], background_rois_row))

            if signal_centre[0] > round(dcm.Columns / 2):
                # phantom is right half of image need 4 ROIs evenly spaced from 0->background_roi[0]
                gap = round(background_rois[0][0] / 4)
                background_rois = [
                    (background_rois[0][0] - i * gap, background_rois_row)
                    for i in range(4)
                ]
            else:
                # phantom is left half of image need 4 ROIs evenly spaced from background_roi[0]->end
                gap = round((dcm.Columns - background_rois[0][0]) / 4)
                background_rois = [
                    (background_rois[0][0] + i * gap, background_rois_row)
                    for i in range(4)
                ]

        else:  # phase encoding is top-down i.e. increases with rows (y-axis)
            if (
                signal_centre[0] < dcm.Columns * 0.5
            ):  # phantom is in left half of image
                background_rois_column = round(
                    dcm.Columns * 0.75
                )  # in the right quadrant
            else:  # phantom is right half of image
                background_rois_column = round(
                    dcm.Columns * 0.25
                )  # in the top quadrant
            background_rois.append((background_rois_column, signal_centre[1]))

            if signal_centre[1] >= round(dcm.Rows / 2):
                # phantom is bottom half of image need 4 ROIs evenly spaced from 0->background_roi[0]
                gap = round(background_rois[0][1] / 4)
                background_rois = [
                    (background_rois_column, background_rois[0][1] - i * gap)
                    for i in range(4)
                ]
            else:  # phantom is top half of image need 3 ROIs evenly spaced from background_roi[0]->end
                gap = round((dcm.Columns - background_rois[0][1]) / 4)
                background_rois = [
                    (background_rois_column, background_rois[0][1] + i * gap)
                    for i in range(4)
                ]

        return background_rois

    def get_background_slices(self, background_rois, slice_radius=5):
        """_summary_

        Args:
            background_rois (list): list of pixel arrays (np.array)
            slice_radius (int, optional): _description_. Defaults to 5.

        Returns:
            list: _description_
        """
        slices = [
            (
                np.array(
                    range(roi[0] - slice_radius, roi[0] + slice_radius),
                    dtype=np.intp,
                )[:, np.newaxis],
                np.array(
                    range(roi[1] - slice_radius, roi[1] + slice_radius),
                    dtype=np.intp,
                ),
            )
            for roi in background_rois
        ]

        return slices

    def get_eligible_area(self, signal_bounding_box, dcm, slice_radius=5):
        """Get pixel array within ROI from image

        Args:
            signal_bounding_box (_type_): _description_
            dcm (pydicom.Dataset): DICOM image object
            slice_radius (int, optional): _description_. Defaults to 5.

        Returns:
            tuple of lists: corresponding to eligible_columns, eligible_rows
        """
        left_column, right_column, upper_row, lower_row = signal_bounding_box

        # take into account when phantom is off edge of image
        lower_row = min(dcm.Rows - slice_radius, lower_row)
        upper_row = max(slice_radius, upper_row)
        right_column = min(dcm.Columns - slice_radius, right_column)
        left_column = max(slice_radius, left_column)

        padding_from_box = 30  # pixels

        if get_pe_direction(dcm) == "ROW":
            if left_column < dcm.Columns / 2:
                # signal is in left half
                eligible_columns = range(
                    right_column + padding_from_box, dcm.Columns - slice_radius
                )
                eligible_rows = range(upper_row, lower_row)

            else:
                # signal is in right half
                eligible_columns = range(
                    slice_radius, left_column - padding_from_box
                )
                eligible_rows = range(upper_row, lower_row)

        else:
            if upper_row < dcm.Rows / 2:
                # signal is in top half
                eligible_rows = range(
                    lower_row + padding_from_box, dcm.Rows - slice_radius
                )
                eligible_columns = range(left_column, right_column)

            else:
                # signal is in bottom half
                eligible_rows = range(
                    slice_radius, upper_row - padding_from_box
                )
                eligible_columns = range(left_column, right_column)

        return eligible_columns, eligible_rows

    def get_ghost_slice(self, signal_bounding_box, dcm, slice_radius=5):
        """Get array of pixel values wihtin bounding box of the ghost slice

        Args:
            signal_bounding_box (tuple or list): _description_
            dcm (pydicom.Dataset): DICOM image object
            slice_radius (int, optional): _description_. Defaults to 5.

        Returns:
            tuple of np.array: _description_
        """
        eligible_area = self.get_eligible_area(
            signal_bounding_box, dcm, slice_radius=slice_radius
        )
        ghost_slice = (
            np.array(
                range(min(eligible_area[1]), max(eligible_area[1])),
                dtype=np.intp,
            )[:, np.newaxis],
            np.array(range(min(eligible_area[0]), max(eligible_area[0]))),
        )
        return ghost_slice

    def get_ghosting(self, dcm) -> float:
        """Calculate ghosting percentage

        Args:
            dcm (pydicom.Dataset): DICOM image object

        Returns:
            float: percentage ghosting across eligible area
        """
        bbox = self.get_signal_bounding_box(dcm.pixel_array)

        x, y = get_pixel_size(dcm)  # assume square pixels i.e. x=y
        # ROIs need to be 10mmx10mm
        slice_radius = int(10 // (2 * x))

        signal_centre = [(bbox[0] + bbox[1]) // 2, (bbox[2] + bbox[3]) // 2]
        background_rois = self.get_background_rois(dcm, signal_centre)
        ghost_col, ghost_row = self.get_ghost_slice(
            bbox, dcm, slice_radius=slice_radius
        )
        ghost = dcm.pixel_array[(ghost_col, ghost_row)]
        signal_col, signal_row = self.get_signal_slice(
            bbox, slice_radius=slice_radius
        )
        phantom = dcm.pixel_array[(signal_row, signal_col)]

        noise = np.concatenate(
            [
                dcm.pixel_array[(row, col)]
                for col, row in self.get_background_slices(
                    background_rois, slice_radius=slice_radius
                )
            ]
        )

        eligible_area = self.get_eligible_area(
            bbox, dcm, slice_radius=slice_radius
        )

        ghosting = self.calculate_ghost_intensity(ghost, phantom, noise)

        if self.report:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()
            x1, x2, y1, y2 = bbox

            img = dcm.pixel_array
            img = img.astype("float64")
            # print('this is img',img)
            img *= 255.0 / img.max()
            img = rescale_to_byte(dcm.pixel_array)
            # img = hazenlib.utils.rescale_to_byte(dcm.pixel_array)
            img = cv.rectangle(img.copy(), (x1, y1), (x2, y2), (255, 0, 0), 1)

            for roi in background_rois:
                #  slice_size = 10
                x1 = roi[0] - 5
                y1 = roi[1] - 5
                x2 = roi[0] + 5
                y2 = roi[1] + 5
                img = cv.rectangle(
                    img.copy(), (x1, y1), (x2, y2), (255, 0, 0), 1
                )

            x1 = ghost_row.min()
            y1 = ghost_col.min()
            x2 = ghost_row.max()
            y2 = ghost_col.max()
            img = cv.rectangle(img.copy(), (x1, y1), (x2, y2), (255, 0, 0), 1)

            x1 = min(eligible_area[0])
            y1 = min(eligible_area[1])
            x2 = max(eligible_area[0])
            y2 = max(eligible_area[1])
            img = cv.rectangle(img.copy(), (x1, y1), (x2, y2), (255, 0, 0), 1)

            ax.imshow(img)
            # fig.savefig(f'{self.report_path}.png')
            img_path = os.path.realpath(
                os.path.join(
                    self.report_path,
                    f"{self.img_desc(dcm, properties=['SeriesDescription', 'EchoTime', 'NumberOfAverages'])}.png",
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return ghosting



"""
Local Otsu thresholding
http://scikit-image.org/docs/0.11.x/auto_examples/plot_local_otsu.html

"""

import copy
import os

import cv2 as cv
import hazenlib.exceptions
import hazenlib.utils
import numpy as np
import pydicom
from hazenlib.HazenTask import HazenTask
from hazenlib.types import Measurement, Result
from skimage import filters, measure


class SlicePosition(HazenTask):
    """Slice position measurement class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(self, *, verbose: bool = False, **kwargs):
        # Whether the position of each of the 40 slices to be included in the results
        self.verbose = verbose
        super().__init__(**kwargs)

    def run(self) -> Result:
        """Main function for performing slice position measurement

        Notes:
            expects an input of 60 images (as a list), of which first and last 10 are discarded

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        if len(self.dcm_list) != 60:
            raise Exception("Need 60 DICOM")

        slice_data = copy.deepcopy(self.dcm_list)
        slice_data.sort(
            key=lambda x: x.SliceLocation
        )  # sort by slice location
        truncated_data = slice_data[10:50]  # ignore first and last 10 dicom

        results = self.init_result_dict()
        results.files = self.img_desc(truncated_data[18])

        try:
            position_errors = self.slice_position_error(truncated_data)
            if self.verbose:
                rounded_positions = [round(pos, 3) for pos in position_errors]
                results.add_measurement(
                    Measurement(
                        name="SlicePosition",
                        type="fitted",
                        subtype="rounded",
                        unit="mm",
                        value=rounded_positions,
                        visibility="intermediate",
                    ),
                )

            # Round calculated values to the appropriate decimal places
            max_pos = round(np.max(position_errors), 2)
            avg_pos = round(np.mean(position_errors), 2)

            results.add_measurement(
                Measurement(
                    name="SlicePosition",
                    type="measured",
                    subtype="maximum",
                    unit="mm",
                    value=max_pos,
                ),
            )
            results.add_measurement(
                Measurement(
                    name="SlicePosition",
                    type="measured",
                    subtype="average",
                    unit="mm",
                    value=avg_pos,
                ),
            )

        except Exception:
            raise

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def get_rod_rotation(self, x_pos: list, y_pos: list) -> float:
        """
        Determine the in-plane rotation i.e. the x-position of the rods should not vary with y-position. If they do it's
        because the phantom is rotated in-plane.

        We can determine the angle of in-plane rotation by plotting the x-position against the y-position. We then fit a
        straight line through the points.
        arctan (gradient) is the angle of rotation.

        If y=c+mx, we can formulate the straight line fit matrix problem, y=X*beta where y is the x-position (because if I
        set y to be the y-position the fit isn't very good because the x-position hardly varies ), X is the two column
        design matrix, the first  column is a constant and the second column are the y-positions.

        Args:
            x_pos (int): x co-ordinate of a rod
            y_pos (int): y co-ordinate of a rod

        Returns:
            float: angle of rotation in degrees

        """
        X = np.array([[i, 1] for i in y_pos])

        m, c = np.linalg.lstsq(X, np.array(x_pos), rcond=None)[0]

        theta = np.arctan(m)
        return theta

    def get_rods_coords(self, dcm: pydicom.Dataset):
        """Determine the coordinates of the rods

        Args:
            dcm (pydicom.Dataset): DICOM image object

        Raises:
            Exception: hazenlib.exceptions.ShapeError

        Returns:
            tuple of int: corresponding to rod coordinates of the left and right rods
        """
        shape_detector = hazenlib.utils.ShapeDetector(arr=dcm.pixel_array)
        try:
            x, y, r = shape_detector.get_shape("circle")

        except hazenlib.exceptions.MultipleShapesError:
            # logger.info(f'Warning: found multiple shapes: {list(shape_detector.shapes.keys())}')
            shape_detector.find_contours()
            shape_detector.detect()
            x, y, r = 0, 0, 0
            for contour in shape_detector.shapes["circle"]:
                (new_x, new_y), new_r = cv.minEnclosingCircle(contour)
                if new_r > r:
                    # logger.info(f"Found bigger circle: {new_x}, {new_y}, {new_r}")
                    x, y, r = new_x, new_y, new_r
                # logger.info(f"Found circle with x={x},y={y},r={r}")

        except hazenlib.exceptions.ShapeError:
            raise

        x, y = int(x), int(y)

        # clip image in xy plane to only include regions where rods could be
        x_window = int(r / 4)
        y_window = int(r * 0.95)

        arr = dcm.pixel_array
        clipped = np.zeros_like(arr)
        clipped[y - y_window : y + y_window, x - x_window : x + x_window] = (
            arr[y - y_window : y + y_window, x - x_window : x + x_window]
        )

        threshold = filters.threshold_otsu(clipped, 2)

        clipped_thresholded = (
            clipped <= threshold
        )  # binarise using otsu threshold

        labels, num = measure.label(clipped_thresholded, return_num=True)
        measured_objects = measure.regionprops(label_image=labels)

        rods = []
        for obj in measured_objects:
            if 5 < obj.bbox_area < 25:
                rods.append(obj)

        if len(rods) != 2:
            raise Exception(f"Found {len(rods)} rods instead of 2.")

        rods.sort(
            key=lambda x: x.centroid[1]
        )  # sort into Left and Right by using second coordinate

        ly, lx = rods[0].centroid
        ry, rx = rods[1].centroid

        # fig = plt.figure(1)
        # plt.imshow(labels)
        # plt.show()

        return lx, ly, rx, ry

    def get_rods(self, data: list):
        """For the whole dataset of 40 DICOMS, record the list of coordinates and
        nominal positions for the left and right rods

        Args:
            data (list): list of pydicom.Dataset image objects

        Returns:
            tuple:
                left_rod and right_rod are a dictionary of lists for y and x coords
                nominal positions is a list calculated from SpacingBetweenSlices
        """
        # TODO: split function so there is no looping
        # TODO: combine this with the function above so rod coords are not recorded again
        left_rod, right_rod = (
            {"x_pos": [], "y_pos": []},
            {"x_pos": [], "y_pos": []},
        )
        nominal_positions = []
        for i, dcm in enumerate(data):
            # print(dcm.SpacingBetweenSlices) # constant
            nominal_positions.append((i + 10) * dcm.SpacingBetweenSlices)

            lx, ly, rx, ry = self.get_rods_coords(dcm)

            left_rod["x_pos"].append(lx)
            left_rod["y_pos"].append(ly)
            right_rod["x_pos"].append(rx)
            right_rod["y_pos"].append(ry)
            # img = dcm.pixel_array
            # cv2.circle(img, (lx, ly), 5, color=(0, 255, 0))
            # cv2.circle(img, (rx, ry), 5, color=(0, 255, 0))
            # fig = plt.figure(1)
            # plt.imshow(img, cmap='gray')
            # plt.show()

        return left_rod, right_rod, nominal_positions

    def correct_rods_for_rotation(
        self, left_rod: dict, right_rod: dict
    ) -> (dict, dict):
        """Update rod coordinates corrected for rotational angle

        Args:
            left_rod (dict): dict of lists for x and y coordinates across 40 slices
            right_rod (dict): dict of lists for x and y coordinates across 40 slices

        Returns:
            tuple of dict: updated lists for x and y coordinates across 40 slices
        """
        r_theta = self.get_rod_rotation(
            x_pos=right_rod["x_pos"], y_pos=right_rod["y_pos"]
        )
        l_theta = self.get_rod_rotation(
            x_pos=left_rod["x_pos"], y_pos=left_rod["y_pos"]
        )
        theta = np.mean([r_theta, l_theta])

        left_rod["x_pos"] = np.subtract(
            np.multiply(np.cos(theta), left_rod["x_pos"]),
            np.multiply(np.sin(theta), left_rod["y_pos"]),
        )

        left_rod["y_pos"] = np.add(
            np.multiply(np.sin(theta), left_rod["x_pos"]),
            np.multiply(np.cos(theta), left_rod["y_pos"]),
        )

        right_rod["x_pos"] = np.subtract(
            np.multiply(np.cos(theta), right_rod["x_pos"]),
            np.multiply(np.sin(theta), right_rod["y_pos"]),
        )

        right_rod["y_pos"] = np.add(
            np.multiply(np.sin(theta), right_rod["x_pos"]),
            np.multiply(np.cos(theta), right_rod["y_pos"]),
        )

        return left_rod, right_rod

    def slice_position_error(self, data: list):
        """Calculate slice position error

        Args:
            data (list): list of pydicom.Dataset image objects

        Returns:
            list: absolute value of difference between nominal and actual positions
        """
        # get rod positions and nominal positions
        left_rod, right_rod, nominal_positions = self.get_rods(data)
        # Correct for phantom rotation
        left_rod, right_rod = self.correct_rods_for_rotation(
            left_rod, right_rod
        )

        fov = hazenlib.utils.get_field_of_view(data[0])

        # x_length_mm = np.subtract(right_rod['x_pos'], left_rod['x_pos']) * fov/data[0].Columns
        y_length_mm = (
            np.subtract(left_rod["y_pos"], right_rod["y_pos"])
            * fov
            / data[0].Columns
        )

        z_length_mm = np.divide(y_length_mm, 2)

        if z_length_mm[0] > z_length_mm[-1]:
            nominal_positions = nominal_positions[::-1]

        # Correct for zero offset
        nominal_positions = [
            x - nominal_positions[18] + z_length_mm[18]
            for x in nominal_positions
        ]
        positions = np.subtract(z_length_mm, nominal_positions)
        distances = [abs(x) for x in positions]

        if self.report:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(2, 1)
            fig.set_size_inches(10, 10)
            ax[0].imshow(data[19].pixel_array, cmap="gray")

            for idx in range(40):
                rods_x = [left_rod["x_pos"][idx], right_rod["x_pos"][idx]]
                rods_y = [left_rod["y_pos"][idx], right_rod["y_pos"][idx]]
                ax[0].scatter(rods_x, rods_y, 20, c="green", marker="+")

            ax[1].scatter(range(10, 50), positions, marker="x")
            ax[1].set_yticks(np.arange(-2.5, 2.5, 0.5))
            plt.xlabel("slice position [slice number]")
            plt.ylabel("Slice position error [mm]")

            img_path = os.path.realpath(
                os.path.join(
                    self.report_path,
                    f"{self.img_desc(self.dcm_list[0])}_slice_position.png",
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

            # fig, ax = plt.subplots(1, 1)
            # for i, pos in enumerate(nominal_positions):
            #     ax.cla()
            #     dcm = self.dcm_list[i+10]
            #     ax.imshow(dcm.pixel_array, cmap='gray')
            #     rods_x = [left_rod["x_pos"][i], right_rod['x_pos'][i]]
            #     rods_y = [left_rod["y_pos"][i], right_rod['y_pos'][i]]
            #     ax.scatter(rods_x, rods_y, 20, c='green', marker='+')
            #
            #     img_path = os.path.realpath(os.path.join(self.report_path, f'{self.key(self.dcm_listdcm_list[0])}_{i}_slice_position.png'))
            #     plt.savefig(img_path)
            #     self.report_files.append(img_path)

        return distances



"""
ACR Spatial Resolution (MTF)

https://www.acraccreditation.org/-/media/acraccreditation/documents/mri/largephantomguidance.pdf

Calculates the effective resolution (MTF50) for slice 1 for the ACR phantom. This is done in accordance with the
methodology described in Section 3 of the following paper:

https://opg.optica.org/oe/fulltext.cfm?uri=oe-22-5-6040&id=281325

WARNING: The phantom must be slanted for valid results to be produced. This test is not within the scope of ACR
guidance.

This script first identifies the rotation angle of the ACR phantom using slice 1. It provides a warning if the
slanted angle is less than 3 degrees.

The location of the ramps within the slice thickness are identified and a square ROI is selected around the anterior
edge of the slice thickness insert.

A rudimentary edge response function is generated based on the edge within the ROI to provide initialisation values for
the 2D normal cumulative distribution fit of the ROI.

The edge is then super-sampled in the direction of the bright-dark transition of the edge and binned at right angles
based on the edge slope determined from the 2D Normal CDF fit of the ROI to obtain the edge response function.

This super-sampled ERF is then fitted using a weighted sigmoid function. The raw data and this fit are then used to
determine the LSF and the subsequent MTF. The MTF50 for both raw and fitted data are reported.

The results are also visualised.

Created by Yassine Azma
yassine.azma@rmh.nhs.uk

22/02/2023
"""

import os
import sys
import traceback
import numpy as np

import cv2
import scipy
import skimage.morphology
import skimage.measure

from hazenlib.HazenTask import HazenTask
from hazenlib.ACRObject import ACRObject
from hazenlib.logger import logger


class ACRSpatialResolution(HazenTask):
    """Spatial resolution measurement class for DICOM images of the ACR phantom

    Inherits from HazenTask class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ACR_obj = ACRObject(self.dcm_list)

    def run(self) -> dict:
        """Main function for performing spatial resolution measurement
        using slice 1 from the ACR phantom image set

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        # Identify relevant slices
        mtf_dcm = self.ACR_obj.slice_stack[0]

        rot_ang = self.ACR_obj.determine_rotation(mtf_dcm.pixel_array)
        if np.abs(rot_ang) < 3:
            logger.warning(
                f"The estimated rotation angle of the ACR phantom is {np.round(rot_ang, 3)} degrees, which "
                f"is less than the recommended 3 degrees. Results will be unreliable!"
            )

        # Initialise results dictionary
        results = self.init_result_dict()
        results.files = [self.img_desc(mtf_dcm)]

        try:
            raw_res, fitted_res = self.get_mtf50(mtf_dcm)
            results["measurement"] = {
                "estimated rotation angle": round(rot_ang, 2),
                "raw mtf50": round(raw_res, 2),
                "fitted mtf50": round(fitted_res, 2),
            }
        except Exception as e:
            print(
                f"Could not calculate the spatial resolution for {self.img_desc(mtf_dcm)} because of : {e}"
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results["report_image"] = self.report_files

        return results

    def y_position_for_ramp(self, img, cxy):
        """Identify the y coordinate of the ramp

        Args:
            img (np.ndarray): dcm.pixelarray
            cxy (tuple): x,y coordinates of the object centre

        Returns:
            float: y coordinate of the ramp min
        """
        investigate_region = int(np.ceil(5.5 / self.ACR_obj.dy).item())

        if np.mod(investigate_region, 2) == 0:
            investigate_region = investigate_region + 1

        line_profile_y = skimage.measure.profile_line(
            img,
            (cxy[1] - 2 * investigate_region, cxy[0]),
            (cxy[1] + 2 * investigate_region, cxy[1]),
            mode="constant",
        ).flatten()

        abs_diff_y_profile = np.absolute(np.diff(line_profile_y))
        y_peaks = scipy.signal.find_peaks(abs_diff_y_profile, height=1)
        pk_heights = y_peaks[1]["peak_heights"]
        pk_ind = y_peaks[0]
        highest_y_peaks = pk_ind[(-pk_heights).argsort()[:2]]
        y_locs = highest_y_peaks - 1

        height_pts = cxy[1] - 2 * investigate_region - 1 + y_locs

        y = np.min(height_pts) + 2

        return y

    def crop_image(self, img, x, y, width):
        """Return a rectangular subset of a pixel array

        Args:
            img (np.ndarray): dcm.pixelarray
            x (int): x coordinate of centre
            y (int): y coordinate of centre
            width (int): size of the array top subset

        Returns:
            np.ndarray: subset of a pixel array with given width
        """
        crop_x, crop_y = (
            (x - width // 2, x + width // 2),
            (
                y - width // 2,
                y + width // 2,
            ),
        )
        crop_img = img[crop_y[0] : crop_y[1], crop_x[0] : crop_x[1]]

        return crop_img

    def get_edge_type(self, crop_img):
        """Determine direction of ramp edge

        Args:
            crop_img (np.ndarray): cropped pixel array ~ subset of the image

        Returns:
            tuple of string: vertical/horizontal and up/down or left/rigtward
        """
        edge_sum_rows = np.sum(crop_img, axis=1).astype(np.int_)
        edge_sum_cols = np.sum(crop_img, axis=0).astype(np.int_)

        _, pk_rows_height = self.ACR_obj.find_n_highest_peaks(
            np.abs(np.diff(edge_sum_rows)), 1
        )
        _, pk_cols_height = self.ACR_obj.find_n_highest_peaks(
            np.abs(np.diff(edge_sum_cols)), 1
        )

        edge_type = (
            "vertical" if pk_rows_height > pk_cols_height else "horizontal"
        )

        thresh_roi_crop = crop_img > 0.6 * np.max(crop_img)
        edge_dir = (
            np.sum(thresh_roi_crop, axis=0)
            if edge_type == "vertical"
            else np.sum(thresh_roi_crop, axis=1)
        )
        if edge_type == "vertical":
            direction = "downward" if edge_dir[-1] > edge_dir[0] else "upward"
        else:
            direction = (
                "leftward" if edge_dir[-1] > edge_dir[0] else "rightward"
            )

        return edge_type, direction

    def edge_location_for_plot(self, crop_img, edge_type):
        """Determine the location of the edge so it can be visualised

        Args:
            crop_img (np.array): cropped pixel array ~ subset of the image
            edge_type (tuple): vertical/horizontal and up/down or left/rigtward

        Returns:
            np.array: mask array for edge location
        """
        thresh_roi_crop = crop_img > 0.6 * np.max(crop_img)

        naive_lsf = (
            np.abs(np.diff(np.sum(thresh_roi_crop, 1))) > 1
            if edge_type == "vertical"
            else np.abs(np.diff(np.sum(thresh_roi_crop, 0)))
        )
        edge_test = np.diff(np.where(naive_lsf == 0))[0]
        edge_begin = np.where(edge_test > 1)
        edge_loc = np.array(
            [edge_begin, edge_begin + edge_test[edge_begin] - 1]
        ).flatten()

        return edge_loc

    def fit_normcdf_surface(self, crop_img, edge_type, direction):
        """Fit normalised CDF? to surface

        Args:
            crop_img (np.array): cropped pixel array ~ subset of the image
            edge_type (string): vertical/horizontal
            direction (string): up/down or left/rigtward

        Returns:
            tuple of floats: slope, surface
        """
        thresh_roi_crop = crop_img > 0.6 * np.max(crop_img)
        temp_x = np.linspace(
            1, thresh_roi_crop.shape[1], thresh_roi_crop.shape[1]
        )
        temp_y = np.linspace(
            1, thresh_roi_crop.shape[0], thresh_roi_crop.shape[0]
        )
        x, y = np.meshgrid(temp_x, temp_y)

        bright = max(crop_img[thresh_roi_crop])
        dark = 20 + np.min(crop_img[~thresh_roi_crop])

        def func(x, slope, mu, bright, dark):
            """Maths function

            Args:
                x (_type_): _description_
                slope (_type_): _description_
                mu (_type_): _description_
                bright (_type_): _description_
                dark (_type_): _description_

            Returns:
                _type_: _description_
            """
            norm_cdf = (bright - dark) * scipy.stats.norm.cdf(
                x[0], mu + slope * x[1], 0.5
            ) + dark

            return norm_cdf

        sign = 1 if direction in ("downward", "leftward") else -1
        x_data = (
            np.vstack((sign * x.ravel(), y.ravel()))
            if edge_type == "vertical"
            else np.vstack((sign * y.ravel(), x.ravel()))
        )

        popt, pcov = scipy.optimize.curve_fit(
            func,
            x_data,
            crop_img.ravel(),
            p0=[0, 0, bright, dark],
            maxfev=1000,
        )
        surface = func(x_data, popt[0], popt[1], popt[2], popt[3]).reshape(
            crop_img.shape
        )

        slope = (
            1 / popt[0]
            if direction in ("leftward", "upward")
            else -1 / popt[0]
        )

        return slope, surface

    def sample_erf(self, crop_img, slope, edge_type):
        """_summary_

        Args:
            crop_img (np.array): cropped pixel array ~ subset of the image
            slope (float): value of slope of edge
            edge_type (string): vertical/horizontal

        Returns:
            np.array: _description_
        """
        resamp_factor = 8
        if edge_type == "horizontal":
            resample_crop_img = cv2.resize(
                crop_img,
                (crop_img.shape[0] * resamp_factor, crop_img.shape[1]),
            )
        else:
            resample_crop_img = cv2.resize(
                crop_img,
                (crop_img.shape[0], crop_img.shape[1] * resamp_factor),
            )

        mid_loc = [i / 2 for i in resample_crop_img.shape]

        temp_x = np.linspace(
            1, resample_crop_img.shape[1], resample_crop_img.shape[1]
        )
        temp_y = np.linspace(
            1, resample_crop_img.shape[0], resample_crop_img.shape[0]
        )
        x_resample, y_resample = np.meshgrid(temp_x, temp_y)

        erf = []
        n_inside_roi = []
        if edge_type == "horizontal":
            diffY = (y_resample - 1) - mid_loc[0]
            x_prime = x_resample + resamp_factor * diffY * slope

            x_min, x_max = (
                np.min(x_prime).astype(int),
                np.max(x_prime).astype(int),
            )

            for k in range(x_min, x_max):
                erf_val = np.mean(
                    resample_crop_img[(x_prime >= k) & (x_prime < k + 1)]
                )
                erf.append(erf_val)
                number_nonzero = np.count_nonzero(
                    resample_crop_img[(x_prime >= k) & (x_prime < k + 1)]
                )
                n_inside_roi.append(number_nonzero)
        else:
            diffX = (x_resample.shape[0] - 1) - x_resample - mid_loc[1]
            y_prime = np.flipud(y_resample) + resamp_factor * diffX * slope

            y_min, y_max = (
                np.min(y_prime).astype(int),
                np.max(y_prime).astype(int),
            )

            for k in range(y_min, y_max):
                erf_val = np.mean(
                    resample_crop_img[(y_prime >= k) & (y_prime < k + 1)]
                )
                erf.append(erf_val)
                number_nonzero = np.count_nonzero(
                    resample_crop_img[(y_prime >= k) & (y_prime < k + 1)]
                )
                n_inside_roi.append(number_nonzero)

        erf = np.array(erf)
        n_inside_roi = np.array(n_inside_roi)

        erf = erf[n_inside_roi == np.max(n_inside_roi)]

        return erf

    def fit_erf(self, erf):
        """Fit ERF

        Args:
            erf (np.array): _description_

        Returns:
            _type_: _description_
        """
        true_erf = np.diff(erf) > 0.2 * np.max(np.diff(erf))
        turning_points = np.where(true_erf)[0][0], np.where(true_erf)[0][-1]
        weights = 0.5 * np.ones((len(true_erf) + 1))
        weights[turning_points[0] : turning_points[1]] = 1

        def func(x, a, b, c, d, e):
            """Maths function for sigmoid curve equation

            Args:
                x (_type_): _description_
                a (_type_): _description_
                b (_type_): _description_
                c (_type_): _description_
                d (_type_): _description_
                e (_type_): _description_

            Returns:
                _type_: _description_
            """
            sigmoid = a + b / (1 + np.exp(c * (x - d))) ** e

            return sigmoid

        popt, pcov = scipy.optimize.curve_fit(
            func,
            np.arange(1, len(erf) + 1),
            erf,
            sigma=(1 / weights),
            p0=[np.min(erf), np.max(erf), 0, sum(turning_points) / 2, 1],
            maxfev=5000,
        )
        erf_fit = func(
            np.arange(1, len(erf) + 1),
            popt[0],
            popt[1],
            popt[2],
            popt[3],
            popt[4],
        )

        return erf_fit

    def calculate_MTF(self, erf):
        """Calculate MTF

        Args:
            erf (np.array): array of ?

        Returns:
            tuple: freq, lsf, MTF
        """
        lsf = np.diff(erf)
        N = len(lsf)
        n = (
            np.arange(-N / 2, N / 2)
            if N % 2 == 0
            else np.arange(-(N - 1) / 2, (N + 1) / 2)
        )

        resamp_factor = 8
        Fs = 1 / (
            np.sqrt(np.mean(np.square((self.ACR_obj.dx, self.ACR_obj.dy))))
            * (1 / resamp_factor)
        )
        freq = n * Fs / N
        MTF = np.abs(np.fft.fftshift(np.fft.fft(lsf)))
        MTF = MTF / np.max(MTF)

        zero_freq = np.where(freq == 0)[0][0]
        freq = freq[zero_freq:]
        MTF = MTF[zero_freq:]

        return freq, lsf, MTF

    def identify_MTF50(self, freq, MTF):
        """Calculate effective resolution

        Args:
            freq (float or int): _description_
            MTF (float or int): _description_

        Returns:
            float: _description_
        """
        freq_interp = np.arange(0, 1.005, 0.005)
        MTF_interp = np.interp(
            freq_interp, freq, MTF, left=None, right=None, period=None
        )
        equivalent_linepairs = freq_interp[np.argmin(np.abs(MTF_interp - 0.5))]
        eff_res = 1 / (equivalent_linepairs * 2)

        return eff_res

    def get_mtf50(self, dcm):
        """_summary_

        Args:
            dcm (pydicom.Dataset): DICOM image object

        Returns:
            tuple: _description_
        """
        img = dcm.pixel_array
        cxy, _ = self.ACR_obj.find_phantom_center(
            img, self.ACR_obj.dx, self.ACR_obj.dy
        )

        ramp_x = int(cxy[0])
        ramp_y = int(self.y_position_for_ramp(img, cxy))
        width = int(13 * img.shape[0] / 256)
        crop_img = self.crop_image(img, ramp_x, ramp_y, width)
        edge_type, direction = self.get_edge_type(crop_img)
        slope, surface = self.fit_normcdf_surface(
            crop_img, edge_type, direction
        )
        erf = self.sample_erf(crop_img, slope, edge_type)
        erf_fit = self.fit_erf(erf)

        freq, lsf_raw, MTF_raw = self.calculate_MTF(erf)
        _, lsf_fit, MTF_fit = self.calculate_MTF(erf_fit)

        eff_raw_res = self.identify_MTF50(freq, MTF_raw)
        eff_fit_res = self.identify_MTF50(freq, MTF_fit)

        if self.report:
            edge_loc = self.edge_location_for_plot(crop_img, edge_type)
            import matplotlib.pyplot as plt
            import matplotlib.patches as patches

            fig, axes = plt.subplots(5, 1)
            fig.set_size_inches(8, 40)
            fig.tight_layout(pad=4)

            axes[0].imshow(img, interpolation="none")
            rect = patches.Rectangle(
                (ramp_x - width // 2 - 1, ramp_y - width // 2 - 1),
                width,
                width,
                linewidth=1,
                edgecolor="w",
                facecolor="none",
            )
            axes[0].add_patch(rect)
            axes[0].axis("off")
            axes[0].set_title("Segmented Edge")

            axes[1].imshow(crop_img)
            if edge_type == "vertical":
                axes[1].plot(
                    np.arange(0, width - 1),
                    np.mean(edge_loc) - slope * np.arange(0, width - 1),
                    color="r",
                )
            else:
                axes[1].plot(
                    np.mean(edge_loc) + slope * np.arange(0, width - 1),
                    np.arange(0, width - 1),
                    color="r",
                )
            axes[1].axis("off")
            axes[1].set_title("Cropped Edge", fontsize=14)

            axes[2].plot(erf, "rx", ms=5, label="Raw Data")
            axes[2].plot(erf_fit, "k", lw=3, label="Fitted Data")
            axes[2].set_ylabel("Signal Intensity")
            axes[2].set_xlabel("Pixel")
            axes[2].grid()
            axes[2].legend(fancybox="true")
            axes[2].set_title("ERF", fontsize=14)

            axes[3].plot(lsf_raw, "rx", ms=5, label="Raw Data")
            axes[3].plot(lsf_fit, "k", lw=3, label="Fitted Data")
            axes[3].set_ylabel(r"$\Delta$" + " Signal Intensity")
            axes[3].set_xlabel("Pixel")
            axes[3].grid()
            axes[3].legend(fancybox="true")
            axes[3].set_title("LSF", fontsize=14)

            axes[4].plot(
                freq,
                MTF_raw,
                "rx",
                ms=8,
                label=f"Raw Data - {round(eff_raw_res, 2)}mm @ 50%",
            )
            axes[4].plot(
                freq,
                MTF_fit,
                "k",
                lw=3,
                label=f"Weighted Sigmoid Fit of ERF - {round(eff_fit_res, 2)}mm @ 50%",
            )
            axes[4].set_xlabel("Spatial Frequency (lp/mm)")
            axes[4].set_ylabel("Modulation Transfer Ratio")
            axes[4].set_xlim([-0.05, 1])
            axes[4].set_ylim([0, 1.05])
            axes[4].grid()
            axes[4].legend(fancybox="true")
            axes[4].set_title("MTF", fontsize=14)

            img_path = os.path.realpath(
                os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return eff_raw_res, eff_fit_res
